# Copyright (c) 2020 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from typing import NewType, Union, Dict, List

JSON = NewType("JSON", Union[str, int, float, bool, None, Dict[str, 'JSON'], List['JSON']])

UserID = NewType("UserID", str)
EventID = NewType("EventID", str)
RoomID = NewType("RoomID", str)
RoomAlias = NewType("RoomAlias", str)

FilterID = NewType("FilterID", str)

ContentURI = NewType("ContentURI", str)

SyncToken = NewType("SyncToken", str)

DeviceID = NewType("DeviceID", str)
SessionID = NewType("SessionID", str)
SigningKey = NewType("SigningKey", str)  # ed25519
IdentityKey = NewType("IdentityKey", str)  # curve25519



# Copyright (c) 2020 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from typing import Iterator, Optional, TypeVar, Type, Dict, List, Any, cast
from contextlib import contextmanager

from sqlalchemy import Table, Constraint
from sqlalchemy.engine.base import Engine, Connection
from sqlalchemy.engine.result import RowProxy, ResultProxy
from sqlalchemy.sql.base import ImmutableColumnCollection
from sqlalchemy.sql.expression import Select, ClauseElement, and_
from sqlalchemy.ext.declarative import as_declarative, declarative_base
from sqlalchemy.dialects.postgresql import insert as pg_insert

T = TypeVar('T', bound='BaseClass')


class BaseClass:
    """
    Base class for SQLAlchemy models. Provides SQLAlchemy declarative base features and some
    additional utilities.
    """
    __tablename__: str

    db: Engine
    t: Table
    __table__: Table
    c: ImmutableColumnCollection
    column_names: List[str]

    @classmethod
    def bind(cls, db_engine: Engine) -> None:
        cls.db = db_engine
        cls.t = cls.__table__
        cls.c = cls.t.columns
        cls.column_names = cls.c.keys()

    @classmethod
    def copy(cls, bind: Optional[Engine] = None, rebase: Optional[declarative_base] = None) -> Type[T]:
        copy = cast(Type[T], type(cls.__name__, (cls, rebase) if rebase else (cls, ), {}))
        if bind is not None:
            copy.bind(db_engine=bind)
        return copy

    @classmethod
    def _one_or_none(cls: Type[T], rows: ResultProxy) -> Optional[T]:
        """
        Try scanning one row from a ResultProxy and return ``None`` if it fails.

        Args:
            rows: The SQLAlchemy result to scan.

        Returns:
            The scanned object, or ``None`` if there were no rows.
        """
        try:
            return cls.scan(next(rows))
        except StopIteration:
            return None

    @classmethod
    def _all(cls: Type[T], rows: ResultProxy) -> Iterator[T]:
        """
        Scan all rows from a ResultProxy.

        Args:
            rows: The SQLAlchemy result to scan.

        Yields:
            Each row scanned with :meth:`scan`
        """
        for row in rows:
            yield cls.scan(row)

    @classmethod
    def scan(cls: Type[T], row: RowProxy) -> T:
        """
        Read the data from a row into an object.

        Args:
            row: The RowProxy object.

        Returns:
            An object containing the information in the row.
        """
        return cls(**dict(zip(cls.column_names, row)))

    @classmethod
    def _make_simple_select(cls: Type[T], *args: ClauseElement) -> Select:
        """
        Create a simple ``SELECT * FROM table WHERE <args>`` statement.

        Args:
            *args: The WHERE clauses. If there are many elements, they're joined with AND.

        Returns:
            The SQLAlchemy SELECT statement object.
        """
        if len(args) > 1:
            return cls.t.select().where(and_(*args))
        elif len(args) == 1:
            return cls.t.select().where(args[0])
        else:
            return cls.t.select()

    @classmethod
    def _select_all(cls: Type[T], *args: ClauseElement) -> Iterator[T]:
        """
        Select all rows with given conditions. This is intended to be used by table-specific
        select methods.

        Args:
            *args: The WHERE clauses. If there are many elements, they're joined with AND.

        Yields:
            The objects representing the rows read with :meth:`scan`
        """
        yield from cls._all(cls.db.execute(cls._make_simple_select(*args)))

    @classmethod
    def _select_one_or_none(cls: Type[T], *args: ClauseElement) -> T:
        """
        Select one row with given conditions. If no row is found, return ``None``. This is intended
        to be used by table-specific select methods.

        Args:
            *args: The WHERE clauses. If there are many elements, they're joined with AND.

        Returns:
            The object representing the matched row read with :meth:`scan`, or ``None`` if no rows
            matched.
        """
        return cls._one_or_none(cls.db.execute(cls._make_simple_select(*args)))

    def _constraint_to_clause(self, constraint: Constraint) -> ClauseElement:
        return and_(*[column == self.__dict__[name]
                      for name, column in constraint.columns.items()])

    @property
    def _edit_identity(self: T) -> ClauseElement:
        """The SQLAlchemy WHERE clause used for editing and deleting individual rows.
        Usually AND of primary keys."""
        return self._constraint_to_clause(self.t.primary_key)

    def edit(self: T, *, _update_values: bool = True, **values) -> None:
        """
        Edit this row.

        Args:
            _update_values: Whether or not the values in memory should be updated as well as the
                            values in the database.
            **values: The values to change.
        """
        with self.db.begin() as conn:
            conn.execute(self.t.update()
                         .where(self._edit_identity)
                         .values(**values))
        if _update_values:
            for key, value in values.items():
                setattr(self, key, value)

    @contextmanager
    def edit_mode(self: T) -> None:
        """
        Edit this row in a fancy context manager way. This stores the current edit identity, then
        yields to the context manager and finally puts the new values into the row using the old
        edit identity in the WHERE clause.

        >>> class TableClass(Base):
        ...     ...
        >>> db_instance = TableClass(id="something")

        >>> with db_instance.edit_mode():
        ...     db_instance.id = "new_id"
        """
        old_identity = self._edit_identity
        yield old_identity
        with self.db.begin() as conn:
            conn.execute(self.t.update().where(old_identity).values(**self._insert_values))

    def delete(self: T) -> None:
        """Delete this row."""
        with self.db.begin() as conn:
            conn.execute(self.t.delete().where(self._edit_identity))

    @property
    def _insert_values(self: T) -> Dict[str, Any]:
        """Values for inserts. Generally you want all the values in the table."""
        return {column_name: self.__dict__[column_name]
                for column_name in self.column_names
                if column_name in self.__dict__}

    def insert(self) -> None:
        with self.db.begin() as conn:
            conn.execute(self.t.insert().values(**self._insert_values))

    @property
    def _upsert_values(self: T) -> Dict[str, Any]:
        """The values to set when an upsert-insert conflicts and moves to the update part."""
        return self._insert_values

    def _upsert_postgres(self: T, conn: Connection) -> None:
        conn.execute(pg_insert(self.t).values(**self._insert_values)
                     .on_conflict_do_update(constraint=self.t.primary_key,
                                            set_=self._upsert_values))

    def _upsert_sqlite(self: T, conn: Connection) -> None:
        conn.execute(self.t.insert().values(**self._insert_values).prefix_with("OR REPLACE"))

    def _upsert_generic(self: T, conn: Connection):
        conn.execute(self.t.delete().where(self._edit_identity))
        conn.execute(self.t.insert().values(**self._insert_values))

    def upsert(self: T) -> None:
        with self.db.begin() as conn:
            if self.db.dialect.name == "postgresql":
                self._upsert_postgres(conn)
            elif self.db.dialect.name == "sqlite":
                self._upsert_sqlite(conn)
            else:
                self._upsert_generic(conn)

    def __iter__(self):
        for key in self.column_names:
            yield self.__dict__[key]


@as_declarative()
class Base(BaseClass):
    pass



# Copyright (c) 2020 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from typing import Any, Dict, Optional
import json

from ..primitive import JSON
from ..util import SerializableEnum, Serializable


class EventType(Serializable):
    """An event type."""
    by_event_type: Dict[str, 'EventType'] = {}

    class Class(SerializableEnum):
        UNKNOWN = "unknown"
        STATE = "state"
        MESSAGE = "message"
        ACCOUNT_DATA = "account_data"
        EPHEMERAL = "ephemeral"
        TO_DEVICE = "to_device"

    __slots__ = ("t", "t_class")

    def __init__(self, t: str, t_class: Class) -> None:
        object.__setattr__(self, "t", t)
        object.__setattr__(self, "t_class", t_class)
        if t not in self.by_event_type:
            self.by_event_type[t] = self

    def serialize(self) -> JSON:
        return self.t

    @classmethod
    def deserialize(cls, raw: JSON) -> Any:
        return cls.find(raw)

    @classmethod
    def find(cls, t: str, t_class: Optional[Class] = None) -> 'EventType':
        try:
            return cls.by_event_type[t].with_class(t_class)
        except KeyError:
            return EventType(t, t_class=t_class)

    def json(self) -> str:
        return json.dumps(self.serialize())

    @classmethod
    def parse_json(cls, data: str) -> 'EventType':
        return cls.deserialize(json.loads(data))

    def __setattr__(self, *args, **kwargs) -> None:
        raise TypeError("EventTypes are frozen")

    def __delattr__(self, *args, **kwargs) -> None:
        raise TypeError("EventTypes are frozen")

    def __str__(self):
        return self.t

    def __repr__(self):
        return f"EventType(\"{self.t}\", EventType.Class.{self.t_class.name})"

    def __hash__(self):
        return hash(self.t) ^ hash(self.t_class)

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, EventType):
            return False
        return self.t == other.t and self.t_class == other.t_class

    def with_class(self, t_class: Class) -> 'EventType':
        if t_class is None or self.t_class == t_class:
            return self
        return EventType(t=self.t, t_class=t_class)

    @property
    def is_message(self) -> bool:
        """Whether or not the event is a message event."""
        return self.t_class == EventType.Class.MESSAGE

    @property
    def is_state(self) -> bool:
        """Whether or not the event is a state event."""
        return self.t_class == EventType.Class.STATE

    @property
    def is_ephemeral(self) -> bool:
        """Whether or not the event is ephemeral."""
        return self.t_class == EventType.Class.EPHEMERAL

    @property
    def is_account_data(self) -> bool:
        """Whether or not the event is an account data event."""
        return self.t_class == EventType.Class.ACCOUNT_DATA

    @property
    def is_to_device(self) -> bool:
        """Whether or not the event is a to-device event."""
        return self.t_class == EventType.Class.TO_DEVICE


_standard_types = {
    EventType.Class.STATE: {
        "m.room.aliases": "ROOM_ALIASES",
        "m.room.canonical_alias": "ROOM_CANONICAL_ALIAS",
        "m.room.create": "ROOM_CREATE",
        "m.room.join_rules": "ROOM_JOIN_RULES",
        "m.room.member": "ROOM_MEMBER",
        "m.room.power_levels": "ROOM_POWER_LEVELS",
        "m.room.history_visibility": "ROOM_HISTORY_VISIBILITY",
        "m.room.name": "ROOM_NAME",
        "m.room.topic": "ROOM_TOPIC",
        "m.room.avatar": "ROOM_AVATAR",
        "m.room.pinned_events": "ROOM_PINNED_EVENTS",
        "m.room.tombstone": "ROOM_TOMBSTONE",
        "m.room.encryption": "ROOM_ENCRYPTION",
    },
    EventType.Class.MESSAGE: {
        "m.room.redaction": "ROOM_REDACTION",
        "m.room.message": "ROOM_MESSAGE",
        "m.room.encrypted": "ROOM_ENCRYPTED",
        "m.sticker": "STICKER",
        "m.reaction": "REACTION",
    },
    EventType.Class.EPHEMERAL: {
        "m.receipt": "RECEIPT",
        "m.typing": "TYPING",
        "m.presence": "PRESENCE",
    },
    EventType.Class.ACCOUNT_DATA: {
        "m.direct": "DIRECT",
        "m.push_rules": "PUSH_RULES",
        "m.tag": "TAG",
        "m.ignored_user_list": "IGNORED_USER_LIST",
    },
    EventType.Class.TO_DEVICE: {
        "m.room.encrypted": "TO_DEVICE_ENCRYPTED",
        "m.room_key": "ROOM_KEY",
        "m.room_key.withheld": "ROOM_KEY_WITHHELD",
        "org.matrix.room_key.withheld": "ORG_MATRIX_ROOM_KEY_WITHHELD",
        "m.room_key_request": "ROOM_KEY_REQUEST",
        "m.forwarded_room_key": "FORWARDED_ROOM_KEY",
    },
}

for _t_class, _types in _standard_types.items():
    for _t, _name in _types.items():
        _event_type = EventType(t=_t, t_class=_t_class)
        setattr(EventType, _name, _event_type)

EventType.ALL = EventType(t="__ALL__", t_class=EventType.Class.UNKNOWN)
EventType.by_event_type["__ALL__"] = EventType.ALL



# Copyright (c) 2020 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
# Partly based on github.com/Cadair/python-appservice-framework (MIT license)
from typing import Optional, Callable, Awaitable, List, Set
from json import JSONDecodeError
from aiohttp import web
import asyncio
import logging

from mautrix.types import JSON, UserID, RoomAlias, Event, SerializerError

QueryFunc = Callable[[web.Request], Awaitable[Optional[web.Response]]]
HandlerFunc = Callable[[Event], Awaitable]


class AppServiceServerMixin:
    loop: asyncio.AbstractEventLoop
    log: logging.Logger

    hs_token: str

    query_user: Callable[[UserID], JSON]
    query_alias: Callable[[RoomAlias], JSON]

    transactions: Set[str]
    event_handlers: List[HandlerFunc]

    def __init__(self) -> None:
        self.transactions = set()
        self.event_handlers = []

        async def default_query_handler(_):
            return None

        self.query_user = default_query_handler
        self.query_alias = default_query_handler

    def register_routes(self, app: web.Application) -> None:
        app.router.add_route("PUT", "/transactions/{transaction_id}",
                             self._http_handle_transaction)
        app.router.add_route("GET", "/rooms/{alias}", self._http_query_alias)
        app.router.add_route("GET", "/users/{user_id}", self._http_query_user)
        app.router.add_route("PUT", "/_matrix/app/v1/transactions/{transaction_id}",
                             self._http_handle_transaction)
        app.router.add_route("GET", "/_matrix/app/v1/rooms/{alias}", self._http_query_alias)
        app.router.add_route("GET", "/_matrix/app/v1/users/{user_id}", self._http_query_user)

    def _check_token(self, request: web.Request) -> bool:
        try:
            token = request.rel_url.query["access_token"]
        except KeyError:
            try:
                token = request.headers["Authorization"].lstrip("Bearer ")
            except (KeyError, AttributeError):
                return False

        if token != self.hs_token:
            return False

        return True

    async def _http_query_user(self, request: web.Request) -> web.Response:
        if not self._check_token(request):
            return web.Response(status=401)

        try:
            user_id = request.match_info["user_id"]
        except KeyError:
            return web.Response(status=400)

        try:
            response = await self.query_user(user_id)
        except Exception:
            self.log.exception("Exception in user query handler")
            return web.Response(status=500)

        if not response:
            return web.Response(status=404)
        return web.json_response(response)

    async def _http_query_alias(self, request: web.Request) -> web.Response:
        if not self._check_token(request):
            return web.Response(status=401)

        try:
            alias = request.match_info["alias"]
        except KeyError:
            return web.Response(status=400)

        try:
            response = await self.query_alias(alias)
        except Exception:
            self.log.exception("Exception in alias query handler")
            return web.Response(status=500)

        if not response:
            return web.Response(status=404)
        return web.json_response(response)

    async def _http_handle_transaction(self, request: web.Request) -> web.Response:
        if not self._check_token(request):
            return web.Response(status=401)

        transaction_id = request.match_info["transaction_id"]
        if transaction_id in self.transactions:
            return web.json_response({})

        try:
            json = await request.json()
        except JSONDecodeError:
            return web.Response(status=400)

        try:
            events = json["events"]
        except KeyError:
            return web.Response(status=400)

        try:
            await self.handle_transaction(transaction_id, events)
        except Exception:
            self.log.exception("Exception in transaction handler")

        self.transactions.add(transaction_id)

        return web.json_response({})

    @staticmethod
    def _fix_prev_content(raw_event: JSON) -> None:
        try:
            raw_event["unsigned"]["prev_content"]
        except KeyError:
            try:
                raw_event.setdefault("unsigned", {})["prev_content"] = raw_event["prev_content"]
            except KeyError:
                pass

    async def handle_transaction(self, txn_id: str, events: List[JSON]) -> None:
        for raw_event in events:
            try:
                self._fix_prev_content(raw_event)
                event = Event.deserialize(raw_event)
            except SerializerError:
                self.log.exception("Failed to deserialize event %s", raw_event)
            else:
                self.handle_matrix_event(event)

    def handle_matrix_event(self, event: Event) -> None:
        if event.type.is_state and event.state_key is None:
            self.log.debug(f"Not sending {event.event_id} to handlers: expected state_key.")
            return

        async def try_handle(handler_func: HandlerFunc):
            try:
                await handler_func(event)
            except Exception:
                self.log.exception("Exception in Matrix event handler")

        for handler in self.event_handlers:
            self.loop.create_task(try_handle(handler))

    def matrix_event_handler(self, func: HandlerFunc) -> HandlerFunc:
        self.event_handlers.append(func)
        return func



# Copyright (c) 2020 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from typing import Tuple, Union, Optional, Dict, TYPE_CHECKING
import logging
import asyncio
import hashlib
import hmac

from mautrix.types import (Filter, RoomFilter, EventFilter, RoomEventFilter, StateFilter, EventType,
                           RoomID, Serializable, JSON, MessageEvent, EncryptedEvent, StateEvent,
                           EncryptedMegolmEventContent, RequestedKeyInfo, RoomKeyWithheldCode)
from mautrix.appservice import AppService
from mautrix.errors import EncryptionError
from mautrix.client import Client, SyncStore
from mautrix.crypto import (OlmMachine, CryptoStore, StateStore, PgCryptoStore, PickleCryptoStore,
                            DeviceIdentity, RejectKeyShare, TrustState)
from mautrix.util.logging import TraceLogger

from .crypto_state_store import GetPortalFunc, PgCryptoStateStore, SQLCryptoStateStore

try:
    from mautrix.client.state_store.sqlalchemy import UserProfile
except ImportError:
    UserProfile = None

try:
    from mautrix.util.async_db import Database
except ImportError:
    Database = None

if TYPE_CHECKING:
    from mautrix.bridge import Bridge


class EncryptionManager:
    loop: asyncio.AbstractEventLoop
    log: TraceLogger = logging.getLogger("mau.bridge.e2ee")

    client: Client
    crypto: OlmMachine
    crypto_store: Union[CryptoStore, SyncStore]
    crypto_db: Optional[Database]
    state_store: StateStore

    bridge: 'Bridge'
    az: AppService
    login_shared_secret: bytes
    _id_prefix: str
    _id_suffix: str

    sync_task: asyncio.Future
    _share_session_events: Dict[RoomID, asyncio.Event]

    def __init__(self, bridge: 'Bridge', login_shared_secret: str, homeserver_address: str,
                 user_id_prefix: str, user_id_suffix: str, db_url: str,
                 key_sharing_config: Dict[str, bool] = None) -> None:
        self.loop = bridge.loop or asyncio.get_event_loop()
        self.bridge = bridge
        self.az = bridge.az
        self.device_name = bridge.name
        self._id_prefix = user_id_prefix
        self._id_suffix = user_id_suffix
        self.login_shared_secret = login_shared_secret.encode("utf-8")
        self._share_session_events = {}
        self.key_sharing_config = key_sharing_config or {}
        pickle_key = "mautrix.bridge.e2ee"
        if db_url.startswith("postgres://"):
            if not PgCryptoStore or not PgCryptoStateStore:
                raise RuntimeError("Database URL is set to postgres, but asyncpg is not installed")
            self.crypto_db = Database(url=db_url, upgrade_table=PgCryptoStore.upgrade_table,
                                      log=logging.getLogger("mau.crypto.db"), loop=self.loop)
            self.crypto_store = PgCryptoStore("", pickle_key, self.crypto_db)
            self.state_store = PgCryptoStateStore(self.crypto_db, bridge.get_portal)
        elif db_url.startswith("pickle:///"):
            self.crypto_db = None
            self.crypto_store = PickleCryptoStore("", pickle_key, db_url[len("pickle:///"):])
            self.state_store = SQLCryptoStateStore(bridge.get_portal)
        else:
            raise RuntimeError("Unsupported database scheme")
        self.client = Client(base_url=homeserver_address, mxid=self.az.bot_mxid, loop=self.loop,
                             sync_store=self.crypto_store)
        self.crypto = OlmMachine(self.client, self.crypto_store, self.state_store)
        self.crypto.allow_key_share = self.allow_key_share

    async def allow_key_share(self, device: DeviceIdentity, request: RequestedKeyInfo) -> bool:
        require_verification = self.key_sharing_config.get("require_verification", True)
        allow = self.key_sharing_config.get("allow", False)
        if not allow:
            return False
        elif device.trust == TrustState.BLACKLISTED:
            raise RejectKeyShare(f"Rejecting key request from blacklisted device "
                                 f"{device.user_id}/{device.device_id}",
                                 code=RoomKeyWithheldCode.BLACKLISTED,
                                 reason="You have been blacklisted by this device")
        elif device.trust == TrustState.VERIFIED or not require_verification:
            portal = await self.bridge.get_portal(request.room_id)
            if portal is None:
                raise RejectKeyShare(f"Rejecting key request for {request.session_id} from "
                                     f"{device.user_id}/{device.device_id}: room is not a portal",
                                     code=RoomKeyWithheldCode.UNAVAILABLE,
                                     reason="Requested room is not a portal")
            user = await self.bridge.get_user(device.user_id)
            if not await user.is_in_portal(portal):
                raise RejectKeyShare(f"Rejecting key request for {request.session_id} from "
                                     f"{device.user_id}/{device.device_id}: user is not in portal",
                                     code=RoomKeyWithheldCode.UNAUTHORIZED,
                                     reason="You're not in that portal")
            self.log.debug(f"Accepting key request for {request.session_id} from "
                           f"{device.user_id}/{device.device_id}")
            return True
        else:
            raise RejectKeyShare(f"Rejecting key request from unverified device "
                                 f"{device.user_id}/{device.device_id}",
                                 code=RoomKeyWithheldCode.UNVERIFIED,
                                 reason="You have not been verified by this device")

    def _ignore_user(self, user_id: str) -> bool:
        return (user_id.startswith(self._id_prefix) and user_id.endswith(self._id_suffix)
                and user_id != self.az.bot_mxid)

    async def handle_member_event(self, evt: StateEvent) -> None:
        if self._ignore_user(evt.state_key):
            # We don't want to invalidate group sessions because a ghost left or joined
            return
        await self.crypto.handle_member_event(evt)

    async def _share_session_lock(self, room_id: RoomID) -> bool:
        try:
            event = self._share_session_events[room_id]
        except KeyError:
            self._share_session_events[room_id] = asyncio.Event()
            return True
        else:
            await event.wait()
            return False

    async def encrypt(self, room_id: RoomID, event_type: EventType,
                      content: Union[Serializable, JSON]
                      ) -> Tuple[EventType, EncryptedMegolmEventContent]:
        try:
            encrypted = await self.crypto.encrypt_megolm_event(room_id, event_type, content)
        except EncryptionError:
            self.log.debug("Got EncryptionError, sharing group session and trying again")
            if await self._share_session_lock(room_id):
                try:
                    users = await self.az.state_store.get_members_filtered(
                        room_id, self._id_prefix, self._id_suffix, self.az.bot_mxid)
                    await self.crypto.share_group_session(room_id, users)
                finally:
                    self._share_session_events.pop(room_id).set()
            encrypted = await self.crypto.encrypt_megolm_event(room_id, event_type, content)
        return EventType.ROOM_ENCRYPTED, encrypted

    async def decrypt(self, evt: EncryptedEvent) -> MessageEvent:
        decrypted = await self.crypto.decrypt_megolm_event(evt)
        self.log.trace("Decrypted event %s: %s", evt.event_id, decrypted)
        return decrypted

    async def start(self) -> None:
        self.log.debug("Logging in with bridge bot user")
        password = hmac.new(self.login_shared_secret, self.az.bot_mxid.encode("utf-8"),
                            hashlib.sha512).hexdigest()
        if self.crypto_db:
            await self.crypto_db.start()
        await self.crypto_store.open()
        device_id = await self.crypto_store.get_device_id()
        if device_id:
            self.log.debug(f"Found device ID in database: {device_id}")
        await self.client.login(password=password, device_name=self.device_name,
                                device_id=device_id, store_access_token=True, update_hs_url=False)
        await self.crypto.load()
        if not device_id:
            await self.crypto_store.put_device_id(self.client.device_id)
            self.log.debug(f"Logged in with new device ID {self.client.device_id}")
        self.sync_task = self.client.start(self._filter)
        self.log.info("End-to-bridge encryption support is enabled")

    async def stop(self) -> None:
        self.sync_task.cancel()
        await self.crypto_store.close()
        if self.crypto_db:
            await self.crypto_db.stop()

    @property
    def _filter(self) -> Filter:
        all_events = EventType.find("*")
        return Filter(
            account_data=EventFilter(types=[all_events]),
            presence=EventFilter(not_types=[all_events]),
            room=RoomFilter(
                include_leave=False,
                state=StateFilter(not_types=[all_events]),
                timeline=RoomEventFilter(not_types=[all_events]),
                account_data=RoomEventFilter(not_types=[all_events]),
                ephemeral=RoomEventFilter(not_types=[all_events]),
            ),
        )



# Copyright (c) 2020 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
# Partly based on github.com/Cadair/python-appservice-framework (MIT license)
from typing import Optional, Callable, Awaitable, Union, Set, Dict
from aiohttp import web
import aiohttp
import asyncio
import logging

from mautrix.types import JSON, UserID, RoomAlias

from .api import AppServiceAPI, IntentAPI
from .state_store import ASStateStore, FileASStateStore
from .as_handler import AppServiceServerMixin

try:
    import ssl
except ImportError:
    ssl = None

QueryFunc = Callable[[web.Request], Awaitable[Optional[web.Response]]]


class AppService(AppServiceServerMixin):
    """The main AppService container."""

    server: str
    domain: str
    verify_ssl: bool
    tls_cert: str
    tls_key: str
    as_token: str
    hs_token: str
    bot_mxid: UserID
    real_user_content_key: str
    state_store: ASStateStore

    transactions: Set[str]

    query_user: Callable[[UserID], JSON]
    query_alias: Callable[[RoomAlias], JSON]
    ready: bool
    live: bool

    loop: asyncio.AbstractEventLoop
    log: logging.Logger
    app: web.Application
    runner: web.AppRunner

    def __init__(self, server: str, domain: str, as_token: str, hs_token: str, bot_localpart: str,
                 loop: Optional[asyncio.AbstractEventLoop] = None,
                 log: Optional[Union[logging.Logger, str]] = None, verify_ssl: bool = True,
                 tls_cert: Optional[str] = None, tls_key: Optional[str] = None,
                 query_user: QueryFunc = None, query_alias: QueryFunc = None,
                 real_user_content_key: Optional[str] = "net.maunium.appservice.puppet",
                 state_store: ASStateStore = None, aiohttp_params: Dict = None) -> None:
        super().__init__()
        self.server = server
        self.domain = domain
        self.verify_ssl = verify_ssl
        self.tls_cert = tls_cert
        self.tls_key = tls_key
        self.as_token = as_token
        self.hs_token = hs_token
        self.bot_mxid = UserID(f"@{bot_localpart}:{domain}")
        self.real_user_content_key: str = real_user_content_key
        if not state_store:
            file = state_store if isinstance(state_store, str) else "mx-state.json"
            self.state_store = FileASStateStore(path=file, binary=False)
        elif isinstance(state_store, ASStateStore):
            self.state_store = state_store
        else:
            raise ValueError(f"Unsupported state store {type(state_store)}")

        self._http_session = None
        self._intent = None

        self.loop = loop or asyncio.get_event_loop()
        self.log = (logging.getLogger(log) if isinstance(log, str)
                    else log or logging.getLogger("mautrix_appservice"))

        self.query_user = query_user or self.query_user
        self.query_alias = query_alias or self.query_alias
        self.live = True
        self.ready = False

        self.app = web.Application(loop=self.loop, **aiohttp_params if aiohttp_params else {})
        self.app.router.add_route("GET", "/_matrix/mau/live", self._liveness_probe)
        self.app.router.add_route("GET", "/_matrix/mau/ready", self._readiness_probe)
        self.register_routes(self.app)

        self.matrix_event_handler(self.state_store.update_state)

    @property
    def http_session(self) -> aiohttp.ClientSession:
        if self._http_session is None:
            raise AttributeError("the http_session attribute can only be used after starting")
        else:
            return self._http_session

    @property
    def intent(self) -> 'IntentAPI':
        if self._intent is None:
            raise AttributeError("the intent attribute can only be used after starting")
        else:
            return self._intent

    async def __aenter__(self) -> None:
        await self.start()

    async def __aexit__(self) -> None:
        await self.stop()

    async def start(self, host: str = "127.0.0.1", port: int = 8080) -> None:
        await self.state_store.open()
        connector = None
        self.log.debug(f"Starting appservice web server on {host}:{port}")
        if self.server.startswith("https://") and not self.verify_ssl:
            connector = aiohttp.TCPConnector(verify_ssl=False)
        self._http_session = aiohttp.ClientSession(loop=self.loop, connector=connector)
        self._intent = AppServiceAPI(base_url=self.server, bot_mxid=self.bot_mxid, log=self.log,
                                     token=self.as_token, state_store=self.state_store,
                                     real_user_content_key=self.real_user_content_key,
                                     client_session=self._http_session).bot_intent()
        ssl_ctx = None
        if self.tls_cert and self.tls_key:
            ssl_ctx = ssl.create_default_context(ssl.Purpose.CLIENT_AUTH)
            ssl_ctx.load_cert_chain(self.tls_cert, self.tls_key)
        self.runner = web.AppRunner(self.app)
        await self.runner.setup()
        site = web.TCPSite(self.runner, host, port, ssl_context=ssl_ctx)
        await site.start()

    async def stop(self) -> None:
        self.log.debug("Stopping appservice web server")
        await self.runner.cleanup()
        self._intent = None
        await self._http_session.close()
        self._http_session = None
        await self.state_store.close()

    async def _liveness_probe(self, _: web.Request) -> web.Response:
        return web.Response(status=200 if self.live else 500, text="{}")

    async def _readiness_probe(self, _: web.Request) -> web.Response:
        return web.Response(status=200 if self.ready else 500, text="{}")



# Copyright (c) 2020 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
import logging

from asyncpg import Connection

from mautrix.util.async_db.upgrade import UpgradeTable

upgrade_table = UpgradeTable(version_table_name="mx_version", database_name="matrix state cache",
                             log=logging.getLogger("mau.client.db.upgrade"))


@upgrade_table.register(description="Initial revision")
async def upgrade_v1(conn: Connection) -> None:
    await conn.execute("""CREATE TABLE mx_room_state (
        room_id              VARCHAR(255) PRIMARY KEY,
        is_encrypted         BOOLEAN,
        has_full_member_list BOOLEAN,
        encryption           TEXT,
        power_levels         TEXT
    )""")
    await conn.execute("CREATE TYPE membership AS ENUM ('join', 'leave', 'invite', 'ban', 'knock')")
    await conn.execute("""CREATE TABLE mx_user_profile (
        room_id     VARCHAR(255),
        user_id     VARCHAR(255),
        membership  membership NOT NULL,
        displayname VARCHAR(255),
        avatar_url  VARCHAR(255),
        PRIMARY KEY (room_id, user_id)
    )""")



# Copyright (c) 2020 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from typing import Optional, Tuple
import warnings
import asyncio

from mautrix.api import HTTPAPI
from mautrix.types import UserID, DeviceID
from mautrix.util.logging import TraceLogger


class BaseClientAPI:
    """
    BaseClientAPI is the base class for :class:`ClientAPI`. This is separate from the main
    ClientAPI class so that the ClientAPI methods can be split into multiple classes (that
    inherit this class).All those section-specific method classes are inherited by the main
    ClientAPI class to create the full class.
    """

    localpart: str
    domain: str
    _mxid: UserID
    device_id: DeviceID
    api: HTTPAPI
    loop: asyncio.AbstractEventLoop
    log: TraceLogger

    def __init__(self, mxid: UserID = "", device_id: DeviceID = "", api: HTTPAPI = None,
                 loop: Optional[asyncio.AbstractEventLoop] = None, *args, **kwargs) -> None:
        """
        Initialize a ClientAPI. You must either provide the

        Args:
            mxid: The Matrix ID of the user. This is used for things like setting profile metadata.
                Additionally, the homeserver domain is extracted from this string and used for
                setting aliases and such. This can be changed later using `set_mxid`.
            api: The :class:`HTTPAPI` instance to use. You can also pass the ``args`` and ``kwargs``
                to create a HTTPAPI instance rather than creating the instance yourself.``
        """
        if mxid:
            self.mxid = mxid
        else:
            self._mxid = None
            self.localpart = None
            self.domain = None
        self.device_id = device_id
        if loop:
            kwargs["loop"] = loop
        self.api = api or HTTPAPI(*args, **kwargs)
        self.loop = self.api.loop
        self.log = self.api.log

    @classmethod
    def parse_mxid(cls, mxid: UserID) -> Tuple[str, str]:
        warnings.warn("parse_mxid is deprecated, use parse_user_id instead",
                      category=DeprecationWarning)
        return cls.parse_user_id(mxid)

    @classmethod
    def parse_user_id(cls, mxid: UserID) -> Tuple[str, str]:
        """
        Parse the localpart and server name from a Matrix user ID.

        Args:
            mxid: The Matrix user ID.

        Returns:
            A tuple of (localpart, server_name).

        Raises:
            ValueError: if the given user ID is invalid.
        """
        if len(mxid) == 0:
            raise ValueError("User ID is empty")
        elif mxid[0] != "@":
            raise ValueError("User IDs start with @")
        try:
            sep = mxid.index(":")
        except ValueError as e:
            raise ValueError("User ID must contain domain separator") from e
        if sep == len(mxid) - 1:
            raise ValueError("User ID must contain domain")
        return mxid[1:sep], mxid[sep + 1:]

    @property
    def mxid(self) -> UserID:
        return self._mxid

    @mxid.setter
    def mxid(self, mxid: UserID) -> None:
        self.localpart, self.domain = self.parse_user_id(mxid)
        self._mxid = mxid



# Copyright (c) 2020 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from typing import Any, Dict
import asyncio

from mautrix.types import (EncryptedOlmEventContent, EventType, UserID, DeviceID,
                           EncryptionKeyAlgorithm, ToDeviceEventContent)

from .base import BaseOlmMachine, verify_signature_json
from .types import DeviceIdentity, DecryptedOlmEvent, OlmEventKeys
from .sessions import Session

ClaimKeysList = Dict[UserID, Dict[DeviceID, DeviceIdentity]]


class OlmEncryptionMachine(BaseOlmMachine):
    _claim_keys_lock: asyncio.Lock

    def __init__(self):
        self._claim_keys_lock = asyncio.Lock()

    async def _encrypt_olm_event(self, session: Session, recipient: DeviceIdentity,
                                 event_type: EventType, content: Any) -> EncryptedOlmEventContent:
        evt = DecryptedOlmEvent(sender=self.client.mxid, sender_device=self.client.device_id,
                                keys=OlmEventKeys(ed25519=self.account.signing_key),
                                recipient=recipient.user_id,
                                recipient_keys=OlmEventKeys(ed25519=recipient.signing_key),
                                type=event_type, content=content)
        ciphertext = session.encrypt(evt.json())
        await self.crypto_store.update_session(recipient.identity_key, session)
        return EncryptedOlmEventContent(ciphertext={recipient.identity_key: ciphertext},
                                        sender_key=self.account.identity_key)

    async def _create_outbound_sessions(self, users: ClaimKeysList) -> None:
        async with self._claim_keys_lock:
            return await self._create_outbound_sessions_locked(users)

    async def _create_outbound_sessions_locked(self, users: ClaimKeysList) -> None:
        request: Dict[UserID, Dict[DeviceID, EncryptionKeyAlgorithm]] = {}
        for user_id, devices in users.items():
            request[user_id] = {}
            for device_id, identity in devices.items():
                if not await self.crypto_store.has_session(identity.identity_key):
                    request[user_id][device_id] = EncryptionKeyAlgorithm.SIGNED_CURVE25519
            if not request[user_id]:
                del request[user_id]
        if not request:
            return
        keys = await self.client.claim_keys(request)
        for user_id, devices in keys.one_time_keys.items():
            for device_id, one_time_keys in devices.items():
                key_id, one_time_key_data = one_time_keys.popitem()
                one_time_key = one_time_key_data["key"]
                identity = users[user_id][device_id]
                if not verify_signature_json(one_time_key_data, user_id, device_id,
                                             identity.signing_key):
                    self.log.warning(f"Invalid signature for {device_id} of {user_id}")
                else:
                    session = self.account.new_outbound_session(identity.identity_key, one_time_key)
                    await self.crypto_store.add_session(identity.identity_key, session)

    async def send_encrypted_to_device(self, device: DeviceIdentity, event_type: EventType,
                                       content: ToDeviceEventContent) -> None:
        await self._create_outbound_sessions({device.user_id: {device.device_id: device}})
        session = await self.crypto_store.get_latest_session(device.identity_key)
        encrypted_content = await self._encrypt_olm_event(session, device, event_type, content)
        await self.client.send_to_one_device(EventType.TO_DEVICE_ENCRYPTED, device.user_id,
                                             device.device_id, encrypted_content)



# Copyright (c) 2020 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from typing import Dict, Optional, List, Tuple

from mautrix.types import SyncToken, IdentityKey, SessionID, RoomID, EventID, UserID, DeviceID
from mautrix.client.state_store import SyncStore

from .. import OlmAccount, Session, InboundGroupSession, OutboundGroupSession, DeviceIdentity
from .abstract import CryptoStore


class MemoryCryptoStore(CryptoStore, SyncStore):
    _device_id: Optional[DeviceID]
    _sync_token: Optional[SyncToken]
    _account: Optional[OlmAccount]
    _message_indices: Dict[Tuple[IdentityKey, SessionID, int], Tuple[EventID, int]]
    _devices: Dict[UserID, Dict[DeviceID, DeviceIdentity]]
    _olm_sessions: Dict[IdentityKey, List[Session]]
    _inbound_sessions: Dict[Tuple[RoomID, IdentityKey, SessionID], InboundGroupSession]
    _outbound_sessions: Dict[RoomID, OutboundGroupSession]

    def __init__(self, account_id: str, pickle_key: str) -> None:
        self.account_id = account_id
        self.pickle_key = pickle_key

        self._sync_token = None
        self._device_id = None
        self._account = None
        self._message_indices = {}
        self._devices = {}
        self._olm_sessions = {}
        self._inbound_sessions = {}
        self._outbound_sessions = {}

    async def get_device_id(self) -> Optional[DeviceID]:
        return self._device_id

    async def put_device_id(self, device_id: DeviceID) -> None:
        self._device_id = device_id

    async def put_next_batch(self, next_batch: SyncToken) -> None:
        self._sync_token = next_batch

    async def get_next_batch(self) -> SyncToken:
        return self._sync_token

    async def put_account(self, account: OlmAccount) -> None:
        self._account = account

    async def get_account(self) -> OlmAccount:
        return self._account

    async def has_session(self, key: IdentityKey) -> bool:
        return key in self._olm_sessions

    async def get_sessions(self, key: IdentityKey) -> List[Session]:
        return self._olm_sessions.get(key, [])

    async def get_latest_session(self, key: IdentityKey) -> Optional[Session]:
        try:
            return self._olm_sessions[key][-1]
        except (KeyError, IndexError):
            return None

    async def add_session(self, key: IdentityKey, session: Session) -> None:
        self._olm_sessions.setdefault(key, []).append(session)

    async def update_session(self, key: IdentityKey, session: Session) -> None:
        # This is a no-op as the session object is the same one previously added.
        pass

    async def put_group_session(self, room_id: RoomID, sender_key: IdentityKey,
                                session_id: SessionID, session: InboundGroupSession) -> None:
        self._inbound_sessions[(room_id, sender_key, session_id)] = session

    async def get_group_session(self, room_id: RoomID, sender_key: IdentityKey,
                                session_id: SessionID) -> InboundGroupSession:
        return self._inbound_sessions.get((room_id, sender_key, session_id))

    async def has_group_session(self, room_id: RoomID, sender_key: IdentityKey,
                                session_id: SessionID) -> bool:
        return (room_id, sender_key, session_id) in self._inbound_sessions

    async def add_outbound_group_session(self, session: OutboundGroupSession) -> None:
        self._outbound_sessions[session.room_id] = session

    async def update_outbound_group_session(self, session: OutboundGroupSession) -> None:
        # This is a no-op as the session object is the same one previously added.
        pass

    async def get_outbound_group_session(self, room_id: RoomID) -> Optional[OutboundGroupSession]:
        return self._outbound_sessions.get(room_id)

    async def remove_outbound_group_session(self, room_id: RoomID) -> None:
        self._outbound_sessions.pop(room_id, None)

    async def remove_outbound_group_sessions(self, rooms: List[RoomID]) -> None:
        for room_id in rooms:
            self._outbound_sessions.pop(room_id, None)

    async def validate_message_index(self, sender_key: IdentityKey, session_id: SessionID,
                                     event_id: EventID, index: int, timestamp: int) -> bool:
        try:
            return self._message_indices[(sender_key, session_id, index)] == (event_id, timestamp)
        except KeyError:
            self._message_indices[(sender_key, session_id, index)] = (event_id, timestamp)
            return True

    async def get_devices(self, user_id: UserID) -> Optional[Dict[DeviceID, DeviceIdentity]]:
        return self._devices.get(user_id)

    async def get_device(self, user_id: UserID, device_id: DeviceID) -> Optional[DeviceIdentity]:
        return self._devices.get(user_id, {}).get(device_id)

    async def put_devices(self, user_id: UserID, devices: Dict[DeviceID, DeviceIdentity]) -> None:
        self._devices[user_id] = devices

    async def filter_tracked_users(self, users: List[UserID]) -> List[UserID]:
        return [user_id for user_id in users if user_id in self._devices]



# Copyright (c) 2020 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from typing import Dict, Optional, List

from mautrix.types import (UserID, RoomID, Membership, Member, PowerLevelStateEventContent,
                           RoomEncryptionStateEventContent)

from ..abstract import StateStore
from .mx_user_profile import UserProfile
from .mx_room_state import RoomState


class SQLStateStore(StateStore):
    _profile_cache: Dict[RoomID, Dict[UserID, UserProfile]]
    _room_state_cache: Dict[RoomID, RoomState]

    def __init__(self) -> None:
        super().__init__()
        self._profile_cache = {}
        self._room_state_cache = {}

    def _get_user_profile(self, room_id: RoomID, user_id: UserID, create: bool = False
                          ) -> UserProfile:
        if not room_id:
            raise ValueError("room_id is empty")
        elif not user_id:
            raise ValueError("user_id is empty")
        try:
            return self._profile_cache[room_id][user_id]
        except KeyError:
            pass
        if room_id not in self._profile_cache:
            self._profile_cache[room_id] = {}

        profile = UserProfile.get(room_id, user_id)
        if profile:
            self._profile_cache[room_id][user_id] = profile
        elif create:
            profile = UserProfile(room_id=room_id, user_id=user_id, membership=Membership.LEAVE)
            profile.insert()
            self._profile_cache[room_id][user_id] = profile
        return profile

    async def get_member(self, room_id: RoomID, user_id: UserID) -> Optional[Member]:
        profile = self._get_user_profile(room_id, user_id)
        if not profile:
            return None
        return profile.member()

    async def set_member(self, room_id: RoomID, user_id: UserID, member: Member) -> None:
        if not member:
            raise ValueError("member info is empty")
        profile = self._get_user_profile(room_id, user_id, create=True)
        profile.edit(membership=member.membership,
                     displayname=member.displayname or profile.displayname,
                     avatar_url=member.avatar_url or profile.avatar_url)

    async def set_membership(self, room_id: RoomID, user_id: UserID,
                             membership: Membership) -> None:
        await self.set_member(room_id, user_id, Member(membership=membership))

    async def get_members(self, room_id: RoomID) -> Optional[List[UserID]]:
        self._profile_cache[room_id] = {}
        for profile in UserProfile.all_in_room(room_id):
            self._profile_cache[room_id][profile.user_id] = profile
        return [profile.user_id for profile in self._profile_cache[room_id].values()]

    async def get_members_filtered(self, room_id: RoomID, not_prefix: str, not_suffix: str,
                                   not_id: str) -> Optional[List[UserID]]:
        return [profile.user_id for profile
                in UserProfile.all_in_room(room_id, not_suffix, not_prefix, not_id)]

    async def set_members(self, room_id: RoomID, members: Dict[UserID, Member],
                          joined_only: bool = False) -> None:
        UserProfile.bulk_replace(room_id, members, joined_only=joined_only)
        self._get_room_state(room_id, create=True).edit(has_full_member_list=True)
        try:
            del self._profile_cache[room_id]
        except KeyError:
            pass

    async def has_full_member_list(self, room_id: RoomID) -> bool:
        room = self._get_room_state(room_id)
        if not room:
            return False
        return room.has_full_member_list

    async def find_shared_rooms(self, user_id: UserID) -> List[RoomID]:
        return [profile.room_id for profile in UserProfile.find_rooms_with_user(user_id)]

    def _get_room_state(self, room_id: RoomID, create: bool = False) -> RoomState:
        if not room_id:
            raise ValueError("room_id is empty")
        try:
            return self._room_state_cache[room_id]
        except KeyError:
            pass

        room = RoomState.get(room_id)
        if room:
            self._room_state_cache[room_id] = room
        elif create:
            room = RoomState(room_id=room_id)
            room.insert()
            self._room_state_cache[room_id] = room
        return room

    async def has_power_levels_cached(self, room_id: RoomID) -> bool:
        room = self._get_room_state(room_id)
        if not room:
            return False
        return room.has_power_levels

    async def get_power_levels(self, room_id: RoomID) -> Optional[PowerLevelStateEventContent]:
        room = self._get_room_state(room_id)
        if not room:
            return None
        return room.power_levels

    async def set_power_levels(self, room_id: RoomID, content: PowerLevelStateEventContent) -> None:
        if not content:
            raise ValueError("content is empty")
        self._get_room_state(room_id, create=True).edit(power_levels=content)

    async def is_encrypted(self, room_id: RoomID) -> Optional[bool]:
        room = self._get_room_state(room_id)
        if not room:
            return None
        return room.is_encrypted

    async def has_encryption_info_cached(self, room_id: RoomID) -> bool:
        room = self._get_room_state(room_id)
        return room and room.has_encryption_info

    async def get_encryption_info(self, room_id: RoomID) -> Optional[RoomEncryptionStateEventContent]:
        room = self._get_room_state(room_id)
        if not room:
            return None
        return room.encryption

    async def set_encryption_info(self, room_id: RoomID,
                                  content: RoomEncryptionStateEventContent) -> None:
        if not content:
            raise ValueError("content is empty")
        self._get_room_state(room_id, create=True).edit(encryption=content, is_encrypted=True)



# -*- coding: utf-8 -*-

# Copyright 2018 Zil0
# Copyright © 2019 Damir Jelić <poljar@termina.org.uk>
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# This function is part of the matrix-python-sdk and is distributed
# under the APACHE 2.0 licence.

"""Matrix encryption algorithms for file uploads."""

from binascii import Error as BinAsciiError
from typing import Any, Dict, Generator, Iterable, Tuple, Union

import unpaddedbase64
from Crypto import Random
from Crypto.Cipher import AES
from Crypto.Hash import SHA256
from Crypto.Util import Counter

from mautrix.errors import DecryptionError
from mautrix.types import EncryptedFile, JSONWebKey

DataT = Union[bytes, Iterable[bytes]]


def decrypt_attachment(ciphertext: bytes, key: str, hash: str, iv: str) -> bytes:
    """Decrypt an encrypted attachment.

    Args:
        ciphertext: The data to decrypt.
        key: AES_CTR JWK key object.
        hash: Base64 encoded SHA-256 hash of the ciphertext.
        iv: Base64 encoded 16 byte AES-CTR IV.
    Returns:
        The plaintext bytes.
    Raises:
        EncryptionError if the integrity check fails.


    """
    expected_hash = unpaddedbase64.decode_base64(hash)

    h = SHA256.new()
    h.update(ciphertext)

    if h.digest() != expected_hash:
        raise DecryptionError("Mismatched SHA-256 digest.")

    try:
        byte_key: bytes = unpaddedbase64.decode_base64(key)
    except (BinAsciiError, TypeError):
        raise DecryptionError("Error decoding key.")

    try:
        # Drop last 8 bytes, which are 0
        byte_iv: bytes = unpaddedbase64.decode_base64(iv)[:8]
    except (BinAsciiError, TypeError):
        raise DecryptionError("Error decoding initial values.")

    ctr = Counter.new(64, prefix=byte_iv, initial_value=0)

    try:
        cipher = AES.new(byte_key, AES.MODE_CTR, counter=ctr)
    except ValueError as e:
        raise DecryptionError("Failed to create AES cipher") from e

    return cipher.decrypt(ciphertext)


def encrypt_attachment(plaintext: bytes) -> Tuple[bytes, EncryptedFile]:
    """Encrypt data in order to send it as an encrypted attachment.

    Args:
        plaintext: The data to encrypt.

    Returns:
        A tuple with the encrypted bytes and a dict containing the info needed
        to decrypt data. See ``encrypted_attachment_generator()`` for the keys.
    """
    values = list(encrypted_attachment_generator(plaintext))
    return b"".join(values[:-1]), values[-1]


def encrypted_attachment_generator(data: DataT
                                   ) -> Generator[Union[bytes, EncryptedFile], None, None]:
    """Generator to encrypt data in order to send it as an encrypted
    attachment.

    Unlike ``encrypt_attachment()``, this function lazily encrypts and yields
    data, thus it can be used to encrypt large files without fully loading them
    into memory if an iterable of bytes is passed as data.

    Args:
        data: The data to encrypt.

    Yields:
        The encrypted bytes for each chunk of data.
        The last yielded value will be a dict containing the info needed to
        decrypt data. The keys are:
        | key: AES-CTR JWK key object.
        | iv: Base64 encoded 16 byte AES-CTR IV.
        | hashes.sha256: Base64 encoded SHA-256 hash of the ciphertext.
    """

    key = Random.new().read(32)
    # 8 bytes IV
    iv = Random.new().read(8)
    # 8 bytes counter, prefixed by the IV
    ctr = Counter.new(64, prefix=iv, initial_value=0)

    cipher = AES.new(key, AES.MODE_CTR, counter=ctr)
    sha256 = SHA256.new()

    if isinstance(data, bytes):
        data = [data]

    for chunk in data:
        encrypted_chunk = cipher.encrypt(chunk)  # in executor
        sha256.update(encrypted_chunk)  # in executor
        yield encrypted_chunk

    yield _get_decryption_info(key, iv, sha256)


def _get_decryption_info(key: bytes, iv: bytes, sha256: SHA256.SHA256Hash) -> EncryptedFile:
    return EncryptedFile(version="v2", iv=unpaddedbase64.encode_base64(iv + b"\x00" * 8),
                         hashes={"sha256": unpaddedbase64.encode_base64(sha256.digest())},
                         key=JSONWebKey(key_type="oct", algorithm="A256CTR", extractable=True,
                                        key_ops=["encrypt", "decrypt"],
                                        key=unpaddedbase64.encode_base64(key, urlsafe=True)))



# Copyright (c) 2020 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from typing import Optional, List, Any
from abc import ABC, abstractmethod

import attr
from attr import dataclass

from .base import BaseConfig


class ConfigValueError(ValueError):
    def __init__(self, key: str, message: str) -> None:
        super().__init__(f"{key} not configured. {message}" if message else f"{key} not configured")


class ForbiddenKey(str):
    pass


@dataclass
class ForbiddenDefault:
    key: str
    value: Any
    error: Optional[str] = None
    condition: Optional[str] = attr.ib(default=None, kw_only=True)

    def check(self, config: 'BaseConfig') -> bool:
        if self.condition and not config[self.condition]:
            return False
        elif isinstance(self.value, ForbiddenKey):
            return str(self.value) in config[self.key]
        else:
            return config[self.key] == self.value

    @property
    def exception(self) -> ConfigValueError:
        return ConfigValueError(self.key, self.error)


class BaseValidatableConfig(BaseConfig, ABC):
    @property
    @abstractmethod
    def forbidden_defaults(self) -> List[ForbiddenDefault]:
        pass

    def check_default_values(self) -> None:
        for default in self.forbidden_defaults:
            if default.check(self):
                raise default.exception



from .recursive_dict import RecursiveDict
from .base import BaseConfig, BaseMissingError, ConfigUpdateHelper
from .string import BaseStringConfig
from .proxy import BaseProxyConfig
from .file import BaseFileConfig, yaml
from .validation import BaseValidatableConfig, ConfigValueError, ForbiddenDefault, ForbiddenKey



# Copyright (c) 2020 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from typing import Dict, List, Tuple

from html.parser import HTMLParser


class HTMLNode(list):
    tag: str
    text: str
    tail: str
    attrib: Dict[str, str]

    def __repr__(self) -> str:
        return (f"HTMLNode(tag='{self.tag}', attrs={self.attrib}, text='{self.text}', "
                f"tail='{self.tail}', children={list(self)})")

    def __init__(self, tag: str, attrs: List[Tuple[str, str]]) -> None:
        super().__init__()
        self.tag = tag
        self.text = ""
        self.tail = ""
        self.attrib = dict(attrs)


class NodeifyingParser(HTMLParser):
    # From https://www.w3.org/TR/html5/syntax.html#writing-html-documents-elements
    void_tags = ("area", "base", "br", "col", "command", "embed", "hr", "img", "input", "link",
                 "meta", "param", "source", "track", "wbr")

    stack: List[HTMLNode]

    def __init__(self) -> None:
        super().__init__()
        self.stack = [HTMLNode("html", [])]

    def handle_starttag(self, tag: str, attrs: List[Tuple[str, str]]) -> None:
        node = HTMLNode(tag, attrs)
        self.stack[-1].append(node)
        if tag not in self.void_tags:
            self.stack.append(node)

    def handle_startendtag(self, tag, attrs):
        self.stack[-1].append(HTMLNode(tag, attrs))

    def handle_endtag(self, tag: str) -> None:
        if tag == self.stack[-1].tag:
            self.stack.pop()

    def handle_data(self, data: str) -> None:
        if len(self.stack[-1]) > 0:
            self.stack[-1][-1].tail += data
        else:
            self.stack[-1].text += data

    def error(self, message: str) -> None:
        pass


def read_html(data: str) -> HTMLNode:
    parser = NodeifyingParser()
    parser.feed(data)
    return parser.stack[0]



# Copyright (c) 2020 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from typing import Optional, Dict, Any, TYPE_CHECKING
from abc import ABC, abstractmethod
import asyncio
import logging

from mautrix.appservice import AppService, IntentAPI
from mautrix.types import (RoomID, EventID, MessageEventContent, EventType, EncryptionAlgorithm,
                           RoomEncryptionStateEventContent as Encryption)
from mautrix.errors import MatrixError
from mautrix.util.logging import TraceLogger
from mautrix.util.simple_lock import SimpleLock

if TYPE_CHECKING:
    from .user import BaseUser
    from .matrix import BaseMatrixHandler
    from .bridge import Bridge


class BasePortal(ABC):
    log: TraceLogger = logging.getLogger("mau.portal")
    az: AppService
    matrix: 'BaseMatrixHandler'
    bridge: 'Bridge'
    loop: asyncio.AbstractEventLoop
    main_intent: IntentAPI
    mxid: Optional[RoomID]
    encrypted: bool
    backfill_lock: SimpleLock

    @abstractmethod
    async def save(self) -> None:
        pass

    @abstractmethod
    async def handle_matrix_message(self, sender: 'BaseUser', message: MessageEventContent,
                                    event_id: EventID) -> None:
        pass

    async def enable_dm_encryption(self) -> bool:
        try:
            await self.main_intent.invite_user(self.mxid, self.az.bot_mxid)
            await self.az.intent.join_room_by_id(self.mxid)
            await self.main_intent.send_state_event(self.mxid, EventType.ROOM_ENCRYPTION,
                                                    Encryption(EncryptionAlgorithm.MEGOLM_V1))
        except Exception:
            self.log.warning(f"Failed to enable end-to-bridge encryption", exc_info=True)
            return False

        self.encrypted = True
        return True

    async def _send_message(self, intent: IntentAPI, content: MessageEventContent,
                            event_type: EventType = EventType.ROOM_MESSAGE, **kwargs) -> EventID:
        if self.encrypted and self.matrix.e2ee:
            if intent.api.is_real_user:
                content[intent.api.real_user_content_key] = True
            event_type, content = await self.matrix.e2ee.encrypt(self.mxid, event_type, content)
        return await intent.send_message_event(self.mxid, event_type, content, **kwargs)

    @property
    @abstractmethod
    def bridge_info_state_key(self) -> str:
        pass

    @property
    @abstractmethod
    def bridge_info(self) -> Dict[str, Any]:
        pass

    # region Matrix room cleanup

    @abstractmethod
    async def delete(self) -> None:
        pass

    async def cleanup_room(self, intent: IntentAPI, room_id: RoomID,
                           message: str = "Portal deleted", puppets_only: bool = False) -> None:
        try:
            members = await intent.get_room_members(room_id)
        except MatrixError:
            members = []
        for user_id in members:
            puppet = await self.bridge.get_puppet(user_id, create=False)
            if user_id != intent.mxid and (not puppets_only or puppet):
                try:
                    if puppet:
                        await puppet.intent.leave_room(room_id)
                    else:
                        await intent.kick_user(room_id, user_id, message)
                except MatrixError:
                    pass
        try:
            await intent.leave_room(room_id)
        except MatrixError:
            pass

    async def unbridge(self) -> None:
        await self.cleanup_room(self.main_intent, self.mxid, "Room unbridged", puppets_only=True)
        await self.delete()

    async def cleanup_and_delete(self) -> None:
        await self.cleanup_room(self.main_intent, self.mxid)
        await self.delete()

    # endregion



# Copyright (c) 2020 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from typing import Optional
from abc import ABC
import io

from ruamel.yaml import YAML
from ruamel.yaml.comments import CommentedMap

from .base import BaseConfig
from .recursive_dict import RecursiveDict

yaml = YAML()
yaml.indent(4)
yaml.width = 200


class BaseStringConfig(BaseConfig, ABC):
    def __init__(self, data: str, base_data: str) -> None:
        super().__init__()
        self._data = yaml.load(data)
        self._base = RecursiveDict(yaml.load(base_data), CommentedMap)

    def load(self) -> None:
        pass

    def load_base(self) -> Optional[RecursiveDict[CommentedMap]]:
        return self._base

    def save(self) -> str:
        buf = io.StringIO()
        yaml.dump(self._data, buf)
        return buf.getvalue()



# Copyright (c) 2020 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from typing import Dict, Optional, List, Any
from abc import ABC, abstractmethod
import random
import string
import re

from mautrix.util.config import (BaseFileConfig, ConfigUpdateHelper, BaseValidatableConfig,
                                 ForbiddenDefault, ForbiddenKey, yaml)


class BaseBridgeConfig(BaseFileConfig, BaseValidatableConfig, ABC):
    registration_path: str
    _registration: Optional[Dict]
    _check_tokens: bool

    def __init__(self, path: str, registration_path: str, base_path: str) -> None:
        super().__init__(path, base_path)
        self.registration_path = registration_path
        self._registration = None
        self._check_tokens = True

    def save(self) -> None:
        super().save()
        if self._registration and self.registration_path:
            with open(self.registration_path, "w") as stream:
                yaml.dump(self._registration, stream)

    @staticmethod
    def _new_token() -> str:
        return "".join(random.choices(string.ascii_lowercase + string.digits, k=64))

    @property
    def forbidden_defaults(self) -> List[ForbiddenDefault]:
        return [
            ForbiddenDefault("homeserver.address", "https://example.com"),
            ForbiddenDefault("homeserver.domain", "example.com"),
        ] + ([
            ForbiddenDefault("appservice.as_token",
                             "This value is generated when generating the registration",
                             "Did you forget to generate the registration?"),
            ForbiddenDefault("appservice.hs_token",
                             "This value is generated when generating the registration",
                             "Did you forget to generate the registration?"),
        ] if self._check_tokens else [])

    def do_update(self, helper: ConfigUpdateHelper) -> None:
        copy = helper.copy

        copy("homeserver.address")
        copy("homeserver.domain")
        copy("homeserver.verify_ssl")

        copy("appservice.address")
        copy("appservice.hostname")
        copy("appservice.port")
        copy("appservice.max_body_size")

        copy("appservice.tls_cert")
        copy("appservice.tls_key")

        copy("appservice.database")

        copy("appservice.id")
        copy("appservice.bot_username")
        copy("appservice.bot_displayname")
        copy("appservice.bot_avatar")

        copy("appservice.as_token")
        copy("appservice.hs_token")

        copy("logging")

    @property
    def namespaces(self) -> Dict[str, List[Dict[str, Any]]]:
        """
        Generate the user ID and room alias namespace config for the registration as specified in
        https://matrix.org/docs/spec/application_service/r0.1.0.html#application-services
        """
        homeserver = self["homeserver.domain"]

        username_format = self["bridge.username_template"].format(userid=".*")
        alias_format = (self["bridge.alias_template"].format(groupname=".*")
                        if "bridge.alias_template" in self else None)
        group_id = ({"group_id": self["appservice.community_id"]}
                    if self["appservice.community_id"] else {})

        return {
            "users": [{
                "exclusive": True,
                "regex": f"@{username_format}:{homeserver}",
                **group_id,
            }],
            "aliases": [{
                "exclusive": True,
                "regex": f"#{alias_format}:{homeserver}",
            }] if alias_format else []
        }

    def generate_registration(self) -> None:
        self["appservice.as_token"] = self._new_token()
        self["appservice.hs_token"] = self._new_token()

        namespaces = self.namespaces
        bot_username = self["appservice.bot_username"]
        homeserver_domain = self["homeserver.domain"]
        namespaces.setdefault("users", []).append({
            "exclusive": True,
            "regex": re.escape(f"@{bot_username}:{homeserver_domain}"),
        })

        self._registration = {
            "id": self["appservice.id"],
            "as_token": self["appservice.as_token"],
            "hs_token": self["appservice.hs_token"],
            "namespaces": namespaces,
            "url": self["appservice.address"],
            "sender_localpart": self._new_token(),
            "rate_limited": False
        }



from .primitive import (UserID, EventID, RoomID, RoomAlias, FilterID, ContentURI, SyncToken,
                        DeviceID, SessionID, SigningKey, IdentityKey, JSON)
from .filter import Filter, EventFilter, RoomFilter, StateFilter, RoomEventFilter
from .event import (EventType, GenericEvent,

                    RedactionEvent, RedactionEventContent,
                    ReactionEventContent, ReactionEvent,

                    MessageEvent, MessageEventContent, MessageUnsigned, MediaMessageEventContent,
                    LocationMessageEventContent, LocationInfo, RelationType, MessageType, Format,
                    MediaInfo, FileInfo, AudioInfo, VideoInfo, ImageInfo, ThumbnailInfo,
                    TextMessageEventContent, BaseMessageEventContent, RelatesTo, BaseFileInfo,
                    EncryptedFile, JSONWebKey,

                    PowerLevelStateEventContent, Membership, MemberStateEventContent, StateEvent,
                    AliasesStateEventContent, CanonicalAliasStateEventContent, StrippedStateEvent,
                    RoomNameStateEventContent, RoomTopicStateEventContent, StateEventContent,
                    RoomPinnedEventsStateEventContent, StateUnsigned, RoomAvatarStateEventContent,
                    RoomTombstoneStateEventContent, RoomEncryptionStateEventContent,

                    AccountDataEvent, AccountDataEventContent, RoomTagInfo,
                    RoomTagAccountDataEventContent,

                    TypingEventContent, TypingEvent, PresenceEvent, PresenceState,
                    PresenceEventContent, SingleReceiptEventContent, ReceiptEventContent,
                    ReceiptEvent, ReceiptType, EphemeralEvent,

                    EncryptedEvent, EncryptionAlgorithm, EncryptedEventContent,
                    EncryptedOlmEventContent, EncryptedMegolmEventContent, EncryptionKeyAlgorithm,
                    OlmMsgType, OlmCiphertext,

                    ToDeviceEvent, ToDeviceEventContent, RoomKeyWithheldCode,
                    RoomKeyWithheldEventContent, RoomKeyEventContent, KeyRequestAction,
                    RequestedKeyInfo, RoomKeyRequestEventContent, ForwardedRoomKeyEventContent,

                    Event, EventContent)
from .misc import (RoomCreatePreset, RoomDirectoryVisibility, PaginationDirection, RoomAliasInfo,
                   RoomDirectoryResponse, DirectoryPaginationToken, PaginatedMessages,
                   DeviceLists, DeviceOTKCount)
from .users import User, Member, UserSearchResults
from .auth import (LoginType, UserIdentifierType, MatrixUserIdentifier, ThirdPartyIdentifier,
                   PhoneIdentifier, UserIdentifier, LoginResponse, DiscoveryInformation,
                   DiscoveryServer, DiscoveryIntegrations, DiscoveryIntegrationServer,
                   LoginFlow)
from .crypto import UnsignedDeviceInfo, DeviceKeys, ClaimKeysResponse, QueryKeysResponse
from .media import MediaRepoConfig, MXOpenGraph, OpenGraphVideo, OpenGraphImage, OpenGraphAudio
from .util import (Obj, Lst, SerializerError, Serializable, SerializableEnum, SerializableAttrs,
                   serializer, deserializer)



# Copyright (c) 2020 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from typing import Union, NewType

from ..primitive import JSON
from ..util import deserializer, Obj
from .base import EventType, GenericEvent
from .redaction import RedactionEvent, RedactionEventContent
from .message import MessageEvent, MessageEventContent
from .reaction import ReactionEvent, ReactionEventContent
from .encrypted import EncryptedEvent, EncryptedEventContent
from .state import StateEvent, StateEventContent
from .account_data import AccountDataEvent, AccountDataEventContent
from .to_device import ToDeviceEvent, ToDeviceEventContent
from .ephemeral import (ReceiptEvent, PresenceEvent, TypingEvent, ReceiptEventContent,
                        TypingEventContent, EphemeralEvent)

Event = NewType("Event", Union[MessageEvent, ReactionEvent, RedactionEvent, StateEvent, TypingEvent,
                               ReceiptEvent, PresenceEvent, EncryptedEvent, ToDeviceEvent,
                               GenericEvent])

EventContent = Union[MessageEventContent, RedactionEventContent, ReactionEventContent,
                     StateEventContent, AccountDataEventContent, ReceiptEventContent,
                     TypingEventContent, EncryptedEventContent, ToDeviceEventContent, Obj]


@deserializer(Event)
def deserialize_event(data: JSON) -> Event:
    event_type = EventType.find(data.get("type", None))
    if event_type == EventType.ROOM_MESSAGE:
        return MessageEvent.deserialize(data)
    elif event_type == EventType.STICKER:
        data.get("content", {})["msgtype"] = "m.sticker"
        return MessageEvent.deserialize(data)
    elif event_type == EventType.REACTION:
        return ReactionEvent.deserialize(data)
    elif event_type == EventType.ROOM_REDACTION:
        return RedactionEvent.deserialize(data)
    elif event_type == EventType.ROOM_ENCRYPTED:
        return EncryptedEvent.deserialize(data)
    elif event_type.is_to_device:
        return ToDeviceEvent.deserialize(data)
    elif event_type.is_state:
        return StateEvent.deserialize(data)
    elif event_type.is_account_data:
        return AccountDataEvent.deserialize(data)
    elif event_type.is_ephemeral:
        return EphemeralEvent.deserialize(data)
    else:
        return GenericEvent.deserialize(data)


setattr(Event, "deserialize", deserialize_event)



# Copyright (c) 2020 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from typing import Optional, Union, Pattern, Dict, Any, List
from html import escape
import re

from attr import dataclass
import attr

from ..util import ExtensibleEnum, SerializableAttrs, Serializable, Obj, deserializer, no_value
from ..primitive import JSON, ContentURI, EventID
from .base import BaseRoomEvent, BaseUnsigned


# region Message types

class Format(ExtensibleEnum):
    """A message format. Currently only ``org.matrix.custom.html`` is available.
    This will probably be deprecated when extensible events are implemented."""
    HTML: 'Format' = "org.matrix.custom.html"


TEXT_MESSAGE_TYPES = ("m.text", "m.emote", "m.notice")
MEDIA_MESSAGE_TYPES = ("m.image", "m.sticker", "m.video", "m.audio", "m.file")


class MessageType(ExtensibleEnum):
    """A message type."""
    TEXT: 'MessageType' = "m.text"
    EMOTE: 'MessageType' = "m.emote"
    NOTICE: 'MessageType' = "m.notice"
    IMAGE: 'MessageType' = "m.image"
    STICKER: 'MessageType' = "m.sticker"
    VIDEO: 'MessageType' = "m.video"
    AUDIO: 'MessageType' = "m.audio"
    FILE: 'MessageType' = "m.file"
    LOCATION: 'MessageType' = "m.location"

    @property
    def is_text(self) -> bool:
        return self.value in TEXT_MESSAGE_TYPES

    @property
    def is_media(self) -> bool:
        return self.value in MEDIA_MESSAGE_TYPES


# endregion
# region Relations

class InReplyTo:
    def __init__(self, event_id: Optional[EventID] = None,
                 proxy_target: Optional['RelatesTo'] = None) -> None:
        self._event_id = event_id
        self._proxy_target = proxy_target

    @property
    def event_id(self) -> EventID:
        if self._proxy_target:
            return self._proxy_target.event_id
        return self._event_id

    @event_id.setter
    def event_id(self, event_id: EventID) -> None:
        if self._proxy_target:
            self._proxy_target.rel_type = RelationType.REFERENCE
            self._proxy_target.event_id = event_id
        else:
            self._event_id = event_id


class RelationType(ExtensibleEnum):
    ANNOTATION: 'RelationType' = "m.annotation"
    REFERENCE: 'RelationType' = "m.reference"
    REPLACE: 'RelationType' = "m.replace"


@dataclass
class RelatesTo(Serializable):
    """Message relations. Used for reactions, edits and replies."""
    rel_type: RelationType = None
    event_id: Optional[EventID] = None
    key: Optional[str] = None
    _extra: Dict[str, Any] = attr.ib(factory=lambda: {})

    def __bool__(self) -> bool:
        return bool(self.rel_type) and bool(self.event_id)

    @classmethod
    def deserialize(cls, data: JSON) -> Optional['RelatesTo']:
        if not data:
            return None
        try:
            return cls(rel_type=RelationType.deserialize(data.pop("rel_type")),
                       event_id=data.pop("event_id", None), key=data.pop("key", None),
                       extra=data)
        except KeyError:
            pass
        try:
            return cls(rel_type=RelationType.REFERENCE, event_id=data["m.in_reply_to"]["event_id"])
        except KeyError:
            pass
        return None

    def serialize(self) -> JSON:
        if not self:
            return no_value
        data = {
            **self._extra,
            "rel_type": self.rel_type.serialize(),
        }
        if self.rel_type == RelationType.REFERENCE:
            data["m.in_reply_to"] = {
                "event_id": self.event_id
            }
        if self.event_id:
            data["event_id"] = self.event_id
        if self.key:
            data["key"] = self.key
        return data

    def __setitem__(self, key: str, value: Any) -> None:
        if key in ("rel_type", "event_id", "key"):
            return setattr(self, key, value)
        self._extra[key] = value

    def __getitem__(self, item: str) -> None:
        if item in ("rel_type", "event_id", "key"):
            return getattr(self, item)
        return self._extra[item]


# endregion
# region Base event content

class BaseMessageEventContentFuncs:
    """Base class for the contents of all message-type events (currently m.room.message and
    m.sticker). Contains relation helpers."""
    body: str
    _relates_to: Optional[RelatesTo]

    def set_reply(self, reply_to: Union[EventID, 'MessageEvent'], **kwargs) -> None:
        self.relates_to.rel_type = RelationType.REFERENCE
        self.relates_to.event_id = reply_to if isinstance(reply_to, str) else reply_to.event_id

    def set_edit(self, edits: Union[EventID, 'MessageEvent']) -> None:
        self.relates_to.rel_type = RelationType.REPLACE
        self.relates_to.event_id = edits if isinstance(edits, str) else edits.event_id

    def serialize(self) -> JSON:
        data = SerializableAttrs.serialize(self)
        evt = self.get_edit()
        if evt:
            new_content = {**data}
            del new_content["m.relates_to"]
            data["m.new_content"] = new_content
            if "body" in data:
                data["body"] = f"* {data['body']}"
            if "formatted_body" in data:
                data["formatted_body"] = f"* {data['formatted_body']}"
        return data

    @property
    def relates_to(self) -> RelatesTo:
        if self._relates_to is None:
            self._relates_to = RelatesTo()
        return self._relates_to

    @relates_to.setter
    def relates_to(self, relates_to: RelatesTo) -> None:
        self._relates_to = relates_to

    def get_reply_to(self) -> Optional[EventID]:
        if self._relates_to and self._relates_to.rel_type == RelationType.REFERENCE:
            return self._relates_to.event_id
        return None

    def get_edit(self) -> Optional[EventID]:
        if self._relates_to and self._relates_to.rel_type == RelationType.REPLACE:
            return self._relates_to.event_id
        return None

    def trim_reply_fallback(self) -> None:
        pass


@dataclass
class BaseMessageEventContent(BaseMessageEventContentFuncs):
    """Base event content for all m.room.message-type events."""
    msgtype: MessageType = None
    body: str = ""

    external_url: str = None
    _relates_to: Optional[RelatesTo] = attr.ib(default=None, metadata={"json": "m.relates_to"})


# endregion
# region Media info


@dataclass
class JSONWebKey(SerializableAttrs['JSONWebKey']):
    key: str = attr.ib(metadata={"json": "k"})
    algorithm: str = attr.ib(default="A256CTR", metadata={"json": "alg"})
    extractable: bool = attr.ib(default=True, metadata={"json": "ext"})
    key_type: str = attr.ib(default="oct", metadata={"json": "kty"})
    key_ops: List[str] = attr.ib(factory=lambda: ["encrypt", "decrypt"])


@dataclass
class EncryptedFile(SerializableAttrs['EncryptedFile']):
    key: JSONWebKey
    iv: str
    hashes: Dict[str, str]
    url: Optional[ContentURI] = None
    version: str = attr.ib(default="v2", metadata={"json": "v"})


@dataclass
class BaseFileInfo(SerializableAttrs['BaseFileInfo']):
    mimetype: str = None
    size: int = None


@dataclass
class ThumbnailInfo(BaseFileInfo, SerializableAttrs['ThumbnailInfo']):
    """Information about the thumbnail for a document, video, image or location."""
    height: int = attr.ib(default=None, metadata={"json": "h"})
    width: int = attr.ib(default=None, metadata={"json": "w"})
    orientation: int = None


@dataclass
class FileInfo(BaseFileInfo, SerializableAttrs['FileInfo']):
    """Information about a document message."""
    thumbnail_info: Optional[ThumbnailInfo] = None
    thumbnail_file: Optional[EncryptedFile] = None
    thumbnail_url: Optional[ContentURI] = None


@dataclass
class ImageInfo(FileInfo, SerializableAttrs['ImageInfo']):
    """Information about an image message."""
    height: int = attr.ib(default=None, metadata={"json": "h"})
    width: int = attr.ib(default=None, metadata={"json": "w"})
    orientation: int = None


@dataclass
class VideoInfo(ImageInfo, SerializableAttrs['VideoInfo']):
    """Information about a video message."""
    duration: int = None
    orientation: int = None


@dataclass
class AudioInfo(BaseFileInfo, SerializableAttrs['AudioInfo']):
    """Information about an audio message."""
    duration: int = None


MediaInfo = Union[ImageInfo, VideoInfo, AudioInfo, FileInfo, Obj]


@dataclass
class LocationInfo(SerializableAttrs['LocationInfo']):
    """Information about a location message."""
    thumbnail_url: Optional[ContentURI] = None
    thumbnail_info: Optional[ThumbnailInfo] = None
    thumbnail_file: Optional[EncryptedFile] = None


# endregion
# region Event content

@dataclass
class MediaMessageEventContent(BaseMessageEventContent,
                               SerializableAttrs['MediaMessageEventContent']):
    """The content of a media message event (m.image, m.audio, m.video, m.file)"""
    url: Optional[ContentURI] = None
    info: Optional[MediaInfo] = None
    file: Optional[EncryptedFile] = None

    @staticmethod
    @deserializer(MediaInfo)
    def deserialize_info(data: JSON) -> MediaInfo:
        if not isinstance(data, dict):
            return Obj()
        msgtype = data.pop("__mautrix_msgtype", None)
        if msgtype == "m.image" or msgtype == "m.sticker":
            return ImageInfo.deserialize(data)
        elif msgtype == "m.video":
            return VideoInfo.deserialize(data)
        elif msgtype == "m.audio":
            return AudioInfo.deserialize(data)
        elif msgtype == "m.file":
            return FileInfo.deserialize(data)
        else:
            return Obj(**data)


@dataclass
class LocationMessageEventContent(BaseMessageEventContent,
                                  SerializableAttrs['LocationMessageEventContent']):
    geo_uri: str = None
    info: LocationInfo = None


html_reply_fallback_regex: Pattern = re.compile("^<mx-reply>"
                                                r"[\s\S]+?"
                                                "</mx-reply>")


@dataclass
class TextMessageEventContent(BaseMessageEventContent,
                              SerializableAttrs['TextMessageEventContent']):
    """The content of a text message event (m.text, m.notice, m.emote)"""
    format: Format = None
    formatted_body: str = None

    def set_reply(self, reply_to: 'MessageEvent', *, displayname: Optional[str] = None) -> None:
        super().set_reply(reply_to)
        if isinstance(reply_to, str):
            return
        if not self.formatted_body or len(self.formatted_body) == 0 or self.format != Format.HTML:
            self.format = Format.HTML
            self.formatted_body = escape(self.body).replace("\n", "<br/>")
        if isinstance(reply_to, MessageEvent):
            self.formatted_body = reply_to.make_reply_fallback_html(
                displayname) + self.formatted_body
            self.body = reply_to.make_reply_fallback_text(displayname) + self.body

    def trim_reply_fallback(self) -> None:
        if self.get_reply_to():
            self._trim_reply_fallback_text()
            self._trim_reply_fallback_html()

    def _trim_reply_fallback_text(self) -> None:
        if not self.body.startswith("> ") or "\n" not in self.body:
            return
        lines = self.body.split("\n")
        while len(lines) > 0 and lines[0].startswith("> "):
            lines.pop(0)
        # Pop extra newline at end of fallback
        lines.pop(0)
        self.body = "\n".join(lines)

    def _trim_reply_fallback_html(self) -> None:
        if self.formatted_body and self.format == Format.HTML:
            self.formatted_body = html_reply_fallback_regex.sub("", self.formatted_body)


MessageEventContent = Union[TextMessageEventContent, MediaMessageEventContent,
                            LocationMessageEventContent, Obj]


# endregion

@dataclass
class MessageUnsigned(BaseUnsigned, SerializableAttrs['MessageUnsigned']):
    """Unsigned information sent with message events."""
    transaction_id: str = None


html_reply_fallback_format = ("<mx-reply><blockquote>"
                              "<a href='https://matrix.to/#/{room_id}/{event_id}'>In reply to</a> "
                              "<a href='https://matrix.to/#/{sender}'>{displayname}</a><br/>"
                              "{content}"
                              "</blockquote></mx-reply>")

media_reply_fallback_body_map = {
    MessageType.IMAGE: "an image",
    MessageType.STICKER: "a sticker",
    MessageType.AUDIO: "audio",
    MessageType.VIDEO: "a video",
    MessageType.FILE: "a file",
    MessageType.LOCATION: "a location",
}


@dataclass
class MessageEvent(BaseRoomEvent, SerializableAttrs['MessageEvent']):
    """An m.room.message event"""
    content: MessageEventContent
    unsigned: Optional[MessageUnsigned] = None

    @staticmethod
    @deserializer(MessageEventContent)
    def deserialize_content(data: JSON) -> MessageEventContent:
        if not isinstance(data, dict):
            return Obj()
        rel = (data.get("m.relates_to", None) or {})
        if rel.get("rel_type", None) == RelationType.REPLACE.value:
            data = data.get("m.new_content", data)
            data["m.relates_to"] = rel
        msgtype = data.get("msgtype", None)
        if msgtype in TEXT_MESSAGE_TYPES:
            return TextMessageEventContent.deserialize(data)
        elif msgtype in MEDIA_MESSAGE_TYPES:
            data.get("info", {})["__mautrix_msgtype"] = msgtype
            return MediaMessageEventContent.deserialize(data)
        elif msgtype == "m.location":
            return LocationMessageEventContent.deserialize(data)
        else:
            return Obj(**data)

    def make_reply_fallback_html(self, displayname: Optional[str] = None) -> str:
        """Generate the HTML fallback for messages replying to this event."""
        if self.content.msgtype.is_text:
            body = self.content.formatted_body or escape(self.content.body).replace("\n", "<br/")
        else:
            sent_type = media_reply_fallback_body_map[self.content.msgtype] or "a message"
            body = f"sent {sent_type}"
        displayname = escape(displayname) if displayname else self.sender
        return html_reply_fallback_format.format(room_id=self.room_id, event_id=self.event_id,
                                                 sender=self.sender, displayname=displayname,
                                                 content=body)

    def make_reply_fallback_text(self, displayname: Optional[str] = None) -> str:
        """Generate the plaintext fallback for messages replying to this event."""
        if self.content.msgtype.is_text:
            body = self.content.body
        else:
            try:
                body = media_reply_fallback_body_map[self.content.msgtype]
            except KeyError:
                body = "an unknown message type"
        lines = body.strip().split("\n")
        first_line, lines = lines[0], lines[1:]
        fallback_text = f"> <{displayname or self.sender}> {first_line}"
        for line in lines:
            fallback_text += f"\n> {line}"
        fallback_text += "\n\n"
        return fallback_text



# Copyright (c) 2020 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from typing import Any, Dict, List, Union
from collections import defaultdict
from datetime import timedelta
import asyncio
import json

from mautrix.types import (EncryptedMegolmEventContent, EventType, UserID, DeviceID, Serializable,
                           EncryptionAlgorithm, RoomID, EncryptedOlmEventContent, SessionID,
                           RoomKeyWithheldEventContent, RoomKeyWithheldCode, IdentityKey,
                           SigningKey, RelatesTo)
from mautrix.errors import EncryptionError, SessionShareError

from .types import DeviceIdentity, TrustState
from .encrypt_olm import OlmEncryptionMachine
from .device_lists import DeviceListMachine
from .sessions import OutboundGroupSession, InboundGroupSession


class Sentinel:
    pass


already_shared = Sentinel()
key_missing = Sentinel()

SessionEncryptResult = Union[
    type(already_shared),  # already shared
    EncryptedOlmEventContent,  # share successful
    RoomKeyWithheldEventContent,  # won't share
    type(key_missing),  # missing device
]


class MegolmEncryptionMachine(OlmEncryptionMachine, DeviceListMachine):
    _megolm_locks: Dict[RoomID, asyncio.Lock]
    _olm_locks: Dict[IdentityKey, asyncio.Lock]
    _sharing_group_session: Dict[RoomID, asyncio.Event]

    def __init__(self) -> None:
        super().__init__()
        self._megolm_locks = defaultdict(lambda: asyncio.Lock())
        self._olm_locks = defaultdict(lambda: asyncio.Lock())
        self._sharing_group_session = {}

    async def encrypt_megolm_event(self, room_id: RoomID, event_type: EventType, content: Any
                                   ) -> EncryptedMegolmEventContent:
        """
        Encrypt an event for a specific room using Megolm.

        Args:
            room_id: The room to encrypt the message for.
            event_type: The event type.
            content: The event content. Using the content structs in the mautrix.types
                module is recommended.

        Returns:
            The encrypted event content.

        Raises:
            EncryptionError: If a group session has not been shared.
                Use :meth:`share_group_session` to share a group session if this error is raised.
        """
        # The crypto store is async, so we need to make sure only one thing is writing at a time.
        async with self._megolm_locks[room_id]:
            return await self._encrypt_megolm_event(room_id, event_type, content)

    async def _encrypt_megolm_event(self, room_id: RoomID, event_type: EventType, content: Any
                                    ) -> EncryptedMegolmEventContent:
        self.log.debug(f"Encrypting event of type {event_type} for {room_id}")
        session = await self.crypto_store.get_outbound_group_session(room_id)
        if not session:
            raise EncryptionError("No group session created")
        ciphertext = session.encrypt(json.dumps({
            "room_id": room_id,
            "type": event_type.serialize(),
            "content": content.serialize() if isinstance(content, Serializable) else content,
        }))
        try:
            relates_to = content.relates_to
        except AttributeError:
            try:
                relates_to = RelatesTo.deserialize(content["m.relates_to"])
            except KeyError:
                relates_to = None
        await self.crypto_store.update_outbound_group_session(session)
        return EncryptedMegolmEventContent(sender_key=self.account.identity_key,
                                           device_id=self.client.device_id, session_id=session.id,
                                           ciphertext=ciphertext, relates_to=relates_to)

    def is_sharing_group_session(self, room_id: RoomID) -> bool:
        """
        Check if there's a group session being shared for a specific room

        Args:
            room_id: The room ID to check.

        Returns:
            True if a group session share is in progress, False if not
        """
        return room_id in self._sharing_group_session

    async def wait_group_session(self, room_id: RoomID) -> None:
        """
        Wait for a group session to be shared.

        Args:
            room_id: The room ID to wait for.
        """
        try:
            event = self._sharing_group_session[room_id]
            await event.wait()
        except KeyError:
            pass

    async def share_group_session(self, room_id: RoomID, users: List[UserID]) -> None:
        """
        Create a Megolm session for a specific room and share it with the given list of users.

        Note that you must not call this method again before the previous share has finished. You
        should either lock calls yourself, or use :meth:`wait_group_session` to use the built-in
        locking capabilities.

        Args:
            room_id: The room to create the session for.
            users: The list of users to share the session with.

        Raises:
            SessionShareError: If something went wrong while sharing the session.
        """
        if room_id in self._sharing_group_session:
            raise SessionShareError("Already sharing group session for that room")
        self._sharing_group_session[room_id] = asyncio.Event()
        try:
            await self._share_group_session(room_id, users)
        finally:
            self._sharing_group_session.pop(room_id).set()

    async def _share_group_session(self, room_id: RoomID, users: List[UserID]) -> None:
        session = await self.crypto_store.get_outbound_group_session(room_id)
        if session and session.shared and not session.expired:
            raise SessionShareError("Group session has already been shared")
        if not session or session.expired:
            session = await self._new_outbound_group_session(room_id)
        self.log.debug(f"Sharing group session {session.id} for room {room_id} with {users}")

        encryption_info = await self.state_store.get_encryption_info(room_id)
        if encryption_info:
            if encryption_info.algorithm != EncryptionAlgorithm.MEGOLM_V1:
                raise SessionShareError("Room encryption algorithm is not supported")
            session.max_messages = encryption_info.rotation_period_msgs or session.max_messages
            session.max_age = (timedelta(milliseconds=encryption_info.rotation_period_ms)
                               if encryption_info.rotation_period_ms else session.max_age)
            self.log.debug("Got stored encryption state event and configured session to rotate "
                           f"after {session.max_messages} messages or {session.max_age}")

        share_key_msgs = defaultdict(lambda: {})
        withhold_key_msgs = defaultdict(lambda: {})
        missing_sessions: Dict[UserID, Dict[DeviceID, DeviceIdentity]] = defaultdict(lambda: {})
        fetch_keys = []

        for user_id in users:
            devices = await self.crypto_store.get_devices(user_id)
            if devices is None:
                self.log.debug(f"get_devices returned nil for {user_id}, will fetch keys and retry")
                fetch_keys.append(user_id)
            elif len(devices) == 0:
                self.log.debug(f"{user_id} has no devices, skipping")
            else:
                self.log.debug(f"Trying to encrypt group session {session.id} for {user_id}")
                for device_id, device in devices.items():
                    result = await self._encrypt_group_session(session, user_id, device_id, device)
                    if isinstance(result, EncryptedOlmEventContent):
                        share_key_msgs[user_id][device_id] = result
                    elif isinstance(result, RoomKeyWithheldEventContent):
                        withhold_key_msgs[user_id][device_id] = result
                    elif result == key_missing:
                        missing_sessions[user_id][device_id] = device

        if fetch_keys:
            self.log.debug(f"Fetching missing keys for {fetch_keys}")
            fetched_keys = await self._fetch_keys(users, include_untracked=True)
            for user_id, devices in fetched_keys.items():
                missing_sessions[user_id] = devices

        if missing_sessions:
            self.log.debug(f"Creating missing outbound sessions {missing_sessions}")
            await self._create_outbound_sessions(missing_sessions)

        for user_id, devices in missing_sessions.items():
            for device_id, device in devices.items():
                result = await self._encrypt_group_session(session, user_id, device_id, device)
                if isinstance(result, EncryptedOlmEventContent):
                    share_key_msgs[user_id][device_id] = result
                elif isinstance(result, RoomKeyWithheldEventContent):
                    withhold_key_msgs[user_id][device_id] = result
                # We don't care about missing keys at this point

        if len(share_key_msgs) > 0:
            event_count = sum(len(map) for map in share_key_msgs.values())
            self.log.debug(f"Sending {event_count} to-device events to share {session.id}")
            await self.client.send_to_device(EventType.TO_DEVICE_ENCRYPTED, share_key_msgs)
        if len(withhold_key_msgs) > 0:
            event_count = sum(len(map) for map in withhold_key_msgs.values())
            self.log.debug(f"Sending {event_count} to-device events "
                           f"to report {session.id} is withheld")
            await self.client.send_to_device(EventType.ROOM_KEY_WITHHELD, withhold_key_msgs)
            await self.client.send_to_device(EventType.ORG_MATRIX_ROOM_KEY_WITHHELD,
                                             withhold_key_msgs)
        self.log.info(f"Group session {session.id} for {room_id} successfully shared")
        session.shared = True
        await self.crypto_store.add_outbound_group_session(session)

    async def _new_outbound_group_session(self, room_id: RoomID) -> OutboundGroupSession:
        session = OutboundGroupSession(room_id)
        await self._create_group_session(self.account.identity_key, self.account.signing_key,
                                         room_id, session.id, session.session_key)
        return session

    async def _create_group_session(self, sender_key: IdentityKey, signing_key: SigningKey,
                                    room_id: RoomID, session_id: SessionID, session_key: str
                                    ) -> None:
        session = InboundGroupSession(session_key=session_key, signing_key=signing_key,
                                      sender_key=sender_key, room_id=room_id)
        await self.crypto_store.put_group_session(room_id, sender_key, session_id, session)
        self.log.debug(f"Created inbound group session {room_id}/{sender_key}/{session_id}")

    async def _encrypt_group_session(self, session: OutboundGroupSession, user_id: UserID,
                                     device_id: DeviceID, device: DeviceIdentity
                                     ) -> SessionEncryptResult:
        async with self._olm_locks[device.identity_key]:
            return await self._encrypt_group_session_locked(session, user_id, device_id, device)

    async def _encrypt_group_session_locked(self, session: OutboundGroupSession, user_id: UserID,
                                            device_id: DeviceID, device: DeviceIdentity
                                            ) -> SessionEncryptResult:
        key = (user_id, device_id)
        if key in session.users_ignored or key in session.users_shared_with:
            return already_shared
        elif user_id == self.client.mxid and device_id == self.client.device_id:
            session.users_ignored.add(key)
            return already_shared

        if device.trust == TrustState.BLACKLISTED:
            self.log.debug(f"Not encrypting group session {session.id} for {device_id} "
                           f"of {user_id}: device is blacklisted")
            session.users_ignored.add(key)
            return RoomKeyWithheldEventContent(
                room_id=session.room_id, algorithm=EncryptionAlgorithm.MEGOLM_V1,
                session_id=session.id, sender_key=self.account.identity_key,
                code=RoomKeyWithheldCode.BLACKLISTED, reason="Device is blacklisted")
        elif not self.allow_unverified_devices and device.trust == TrustState.UNSET:
            self.log.debug(f"Not encrypting group session {session.id} for {device_id} "
                           f"of {user_id}: device is not verified")
            session.users_ignored.add(key)
            return RoomKeyWithheldEventContent(
                room_id=session.room_id, algorithm=EncryptionAlgorithm.MEGOLM_V1,
                session_id=session.id, sender_key=self.account.identity_key,
                code=RoomKeyWithheldCode.UNVERIFIED, reason="This device does not encrypt "
                                                            "messages for unverified devices")
        device_session = await self.crypto_store.get_latest_session(device.identity_key)
        if not device_session:
            return key_missing
        encrypted = await self._encrypt_olm_event(device_session, device, EventType.ROOM_KEY,
                                                  session.share_content)
        session.users_shared_with.add(key)
        self.log.debug(f"Encrypted group session {session.id} for {device_id} of {user_id}")
        return encrypted



# Copyright (c) 2020 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from typing import Awaitable, Dict, Optional, List

from mautrix.api import Method, Path
from mautrix.errors import MatrixResponseError
from mautrix.types import (JSON, UserID, RoomID, EventID, FilterID, SyncToken, PaginationDirection,
                           StateEvent, EventType, StateEventContent, MessageEventContent, Member,
                           Event, ContentURI, PaginatedMessages, SerializerError, MessageType,
                           RelatesTo, Format, ImageInfo, BaseFileInfo, TextMessageEventContent,
                           MediaMessageEventContent, PresenceState, EventContent, Membership,
                           ReactionEventContent, RelationType, Obj, Serializable)
from mautrix.types.event.state import state_event_content_map
from .base import BaseClientAPI


class EventMethods(BaseClientAPI):
    """
    Methods in section 8 Events of the spec. Includes ``/sync``ing, getting messages and state,
    setting state, sending messages and redacting messages. See also: `Events API reference`_

    .. _Events API reference:
        https://matrix.org/docs/spec/client_server/r0.5.0#events
    """

    # region 8.2 Syncing
    # API reference: https://matrix.org/docs/spec/client_server/r0.4.0.html#syncing

    def sync(self, since: SyncToken = None, timeout: int = 30000, filter_id: FilterID = None,
             full_state: bool = False, set_presence: PresenceState = None) -> Awaitable[JSON]:
        """
        Perform a sync request. See also: `/sync API reference`_

        Args:
            since (str): Optional. A token which specifies where to continue a sync from.
            timeout (int): Optional. The time in milliseconds to wait.
            filter_id (int): A filter ID.
            full_state (bool): Return the full state for every room the user has joined
                Defaults to false.
            set_presence (str): Should the client be marked as "online" or" offline"

        .. _/sync API reference:
            https://matrix.org/docs/spec/client_server/r0.5.0#get-matrix-client-r0-sync
        """
        request = {"timeout": timeout}
        if since:
            request["since"] = str(since)
        if filter_id:
            request["filter"] = str(filter_id)
        if full_state:
            request["full_state"] = "true" if full_state else "false"
        if set_presence:
            request["set_presence"] = str(set_presence)
        return self.api.request(Method.GET, Path.sync, query_params=request)

    # endregion
    # region 8.3 Getting events for a room
    # API reference: https://matrix.org/docs/spec/client_server/r0.4.0.html#getting-events-for-a-room

    async def get_event(self, room_id: RoomID, event_id: EventID) -> Event:
        """
        Get a single event based on ``room_id``/``event_id``. You must have permission to retrieve
        this event e.g. by being a member in the room for this event.
        See also: `/event/{eventId} API reference`_

        Args:
            room_id: The ID of the room the event is in.
            event_id: The event ID to get.

        Returns:
            The event.

        .. _/event/{eventId} API reference:
            https://matrix.org/docs/spec/client_server/r0.5.0#get-matrix-client-r0-rooms-roomid-event-eventid
        """
        content = await self.api.request(Method.GET, Path.rooms[room_id].event[event_id])
        try:
            return Event.deserialize(content)
        except SerializerError as e:
            raise MatrixResponseError("Invalid event in response") from e

    async def get_state_event(self, room_id: RoomID, event_type: EventType,
                              state_key: Optional[str] = None) -> StateEventContent:
        """
        Looks up the contents of a state event in a room. If the user is joined to the room then the
        state is taken from the current state of the room. If the user has left the room then the
        state is taken from the state of the room when they left.
        See also: `GET /state/{eventType}/{stateKey} API reference`_

        Args:
            room_id: The ID of the room to look up the state in.
            event_type: The type of state to look up.
            state_key: The key of the state to look up. Defaults to empty string.

        Returns:
            The state event.

        .. _GET /state/{eventType}/{stateKey} API reference:
            https://matrix.org/docs/spec/client_server/r0.5.0#get-matrix-client-r0-rooms-roomid-state-eventtype-statekey
        """
        content = await self.api.request(Method.GET,
                                         Path.rooms[room_id].state[event_type][state_key])
        try:
            return state_event_content_map[event_type].deserialize(content)
        except KeyError:
            return Obj(**content)
        except SerializerError as e:
            raise MatrixResponseError("Invalid state event in response") from e

    async def get_state(self, room_id: RoomID) -> List[StateEvent]:
        """
        Get the state events for the current state of a room. See also: `/state API reference`_

        Args:
            room_id: The ID of the room to look up the state for.

        Returns:
            A list of state events with the most recent of each event_type/state_key pair.

        .. _/state API reference:
            https://matrix.org/docs/spec/client_server/r0.5.0#get-matrix-client-r0-rooms-roomid-state
        """
        content = await self.api.request(Method.GET, Path.rooms[room_id].state)
        try:
            return [StateEvent.deserialize(event) for event in content]
        except SerializerError as e:
            raise MatrixResponseError("Invalid state events in response") from e

    async def get_members(self, room_id: RoomID) -> List[StateEvent]:
        """
        Get the list of members for a room. See also: `/members API reference`_

        Args:
            room_id: The ID of the room to get the member events for.

        Returns:
            A list of most recent member events for each user.

        .. _/members API reference:
            https://matrix.org/docs/spec/client_server/r0.5.0#get-matrix-client-r0-rooms-roomid-members
        """
        content = await self.api.request(Method.GET, Path.rooms[room_id].members)
        try:
            return [StateEvent.deserialize(event) for event in content["chunk"]]
        except KeyError:
            raise MatrixResponseError("`chunk` not in response.")
        except SerializerError as e:
            raise MatrixResponseError("Invalid state events in response") from e

    async def get_joined_members(self, room_id: RoomID) -> Dict[UserID, Member]:
        """
        Get a user ID -> member info map for a room. The current user must be in the room for it to
        work, unless it is an Application Service in which case any of the AS's users must be in the
        room. This API is primarily for Application Services and should be faster to respond than
        `/members`_ as it can be implemented more efficiently on the server.
        See also: `/joined_members API reference`_

        Args:
            room_id: The ID of the room to get the members of.

        Returns:
            A dictionary from user IDs to Member info objects.

        .. _/joined_members API reference:
            https://matrix.org/docs/spec/client_server/r0.5.0#get-matrix-client-r0-rooms-roomid-joined-members
        .. _/members:
            https://matrix.org/docs/spec/client_server/r0.5.0#get-matrix-client-r0-rooms-roomid-members
        """
        content = await self.api.request(Method.GET, Path.rooms[room_id].joined_members)
        try:
            return {user_id: Member(membership=Membership.JOIN,
                                    displayname=member.get("display_name", ""),
                                    avatar_url=member.get("avatar_url", ""))
                    for user_id, member in content["joined"].items()}
        except KeyError:
            raise MatrixResponseError("`joined` not in response.")
        except SerializerError as e:
            raise MatrixResponseError("Invalid member objects in response") from e

    async def get_messages(self, room_id: RoomID, direction: PaginationDirection,
                           from_token: SyncToken, to_token: Optional[SyncToken] = None,
                           limit: Optional[int] = None, filter_json: Optional[str] = None
                           ) -> PaginatedMessages:
        """
        Get a list of message and state events for a room. Pagination parameters are used to
        paginate history in the room. See also: `/messages API reference`_

        Args:
            room_id: The ID of the room to get events from.
            direction: The direction to return events from.
            from_token: The token to start returning events from. This token can be obtained from a
                ``prev_batch`` token returned for each room by the `sync endpoint`_, or from a
                ``start`` or ``end`` token returned by a previous request to this endpoint.
            to_token: The token to stop returning events at.
            limit: The maximum number of events to return. Defaults to 10.
            filter_json: A JSON RoomEventFilter_ to filter returned events with.

        Returns:

        .. _RoomEventFilter:
            https://matrix.org/docs/spec/client_server/r0.5.0#filtering
        .. _sync endpoint:
            https://matrix.org/docs/spec/client_server/r0.5.0#get-matrix-client-r0-sync
        .. _/messages API reference:
            https://matrix.org/docs/spec/client_server/r0.5.0#get-matrix-client-r0-rooms-roomid-messages
        """
        query_params = {
            "from": from_token,
            "dir": direction.value,
        }
        if to_token:
            query_params["to"] = to_token
        if limit:
            query_params["limit"] = str(limit)
        if filter:
            query_params["filter"] = filter_json
        content = await self.api.request(Method.GET, Path.rooms[room_id].messages,
                                         query_params=query_params)
        try:
            return PaginatedMessages(content["start"], content["end"],
                                     [Event.deserialize(event) for event in content["chunk"]])
        except KeyError:
            if "start" not in content:
                raise MatrixResponseError("`start` not in response.")
            elif "end" not in content:
                raise MatrixResponseError("`start` not in response.")
            raise MatrixResponseError("`content` not in response.")
        except SerializerError as e:
            raise MatrixResponseError("Invalid events in response") from e

    # endregion
    # region 8.4 Sending events to a room
    # API reference: https://matrix.org/docs/spec/client_server/r0.4.0.html#sending-events-to-a-room

    async def send_state_event(self, room_id: RoomID, event_type: EventType,
                               content: StateEventContent, state_key: Optional[str] = "", **kwargs
                               ) -> EventID:
        """
        Send a state event to a room. State events with the same ``room_id``, ``event_type`` and
        ``state_key`` will be overridden.
        See also: `PUT /state/{eventType}/{stateKey} API reference`_

        Args:
            room_id: The ID of the room to set the state in.
            event_type: The type of state to send.
            content: The content to send.
            state_key: The key for the state to send. Defaults to empty string.
            **kwargs: Optional parameters to pass to the :meth:`HTTPAPI.request` method. Used by
                :class:`IntentAPI` to pass the timestamp massaging field to
                :meth:`AppServiceAPI.request`.

        Returns:
            The ID of the event that was sent.

        .. _PUT /state/{eventType}/{stateKey} API reference:
            https://matrix.org/docs/spec/client_server/r0.5.0#put-matrix-client-r0-rooms-roomid-state-eventtype-statekey
        """
        content = content.serialize() if isinstance(content, Serializable) else content
        resp = await self.api.request(Method.PUT, Path.rooms[room_id].state[event_type][state_key],
                                      content, **kwargs)
        try:
            return resp["event_id"]
        except KeyError:
            raise MatrixResponseError("`event_id` not in response.")

    async def send_message_event(self, room_id: RoomID, event_type: EventType,
                                 content: EventContent, **kwargs) -> EventID:
        """
        Send a message event to a room. Message events allow access to historical events and
        pagination, making them suited for "once-off" activity in a room.
        See also: `/send API reference`_

        Args:
            room_id: The ID of the room to send the message to.
            event_type: The type of message to send.
            content: The content to send.
            **kwargs: Optional parameters to pass to the :meth:`HTTPAPI.request` method. Used by
                :class:`IntentAPI` to pass the timestamp massaging field to
                :meth:`AppServiceAPI.request`.

        Returns:
            The ID of the event that was sent.

        .. _/send API reference:
            https://matrix.org/docs/spec/client_server/r0.5.0#put-matrix-client-r0-rooms-roomid-send-eventtype-txnid
        """
        if not room_id:
            raise ValueError("Room ID not given")
        elif not event_type:
            raise ValueError("Event type not given")
        url = Path.rooms[room_id].send[event_type][self.api.get_txn_id()]
        content = content.serialize() if isinstance(content, Serializable) else content
        resp = await self.api.request(Method.PUT, url, content, **kwargs)
        try:
            return resp["event_id"]
        except KeyError:
            raise MatrixResponseError("`event_id` not in response.")

    # region Message send helper functions
    def send_message(self, room_id: RoomID, content: MessageEventContent, **kwargs
                     ) -> Awaitable[EventID]:
        """
        Send a message to a room.

        Args:
            room_id: The ID of the room to send the message to.
            content: The content to send.
            **kwargs: Optional parameters to pass to the :meth:`HTTPAPI.request` method.

        Returns:
            The ID of the event that was sent.
        """
        return self.send_message_event(room_id, EventType.ROOM_MESSAGE, content, **kwargs)

    def react(self, room_id: RoomID, event_id: EventID, key: str, **kwargs) -> Awaitable[EventID]:
        content = ReactionEventContent(relates_to=RelatesTo(rel_type=RelationType.ANNOTATION,
                                                            event_id=event_id, key=key))
        return self.send_message_event(room_id, EventType.REACTION, content, **kwargs)

    def send_text(self, room_id: RoomID, text: str, html: Optional[str] = None,
                  msgtype: MessageType = MessageType.TEXT, relates_to: Optional[RelatesTo] = None,
                  **kwargs) -> Awaitable[EventID]:
        """
        Send a text message to a room.

        Args:
            room_id: The ID of the room to send the message to.
            text: The text to send. If set to None, the given HTML is used instead.
            html: The HTML to send.
            msgtype: The message type to send.
                Defaults to :attr:`MessageType.TEXT` (normal text message)
            relates_to: Message relation metadata used for things like replies.
            **kwargs: Optional parameters to pass to the :meth:`HTTPAPI.request` method.

        Returns:
            The ID of the event that was sent.
        """
        if html:
            if not text:
                text = html
            content = TextMessageEventContent(msgtype=msgtype, body=text,
                                              format=Format.HTML, formatted_body=html)
        else:
            content = TextMessageEventContent(msgtype=msgtype, body=text)
        if relates_to:
            content.relates_to = relates_to
        return self.send_message(room_id, content, **kwargs)

    def send_notice(self, room_id: RoomID, text: str, html: Optional[str] = None,
                    relates_to: Optional[RelatesTo] = None, **kwargs) -> Awaitable[EventID]:
        """
        Send a notice text message to a room. Notices are like normal text messages, but usually
        sent by bots to tell other bots not to react to them. If you're a bot, please send notices
        instead of normal text, unless there is a reason to do something else.

        Args:
            room_id: The ID of the room to send the message to.
            text: The text to send. If set to None, the given HTML is used instead.
            html: The HTML to send.
            msgtype: The message type to send.
                Defaults to :attr:`MessageType.TEXT` (normal text message)
            relates_to: Message relation metadata used for things like replies.
            **kwargs: Optional parameters to pass to the :meth:`HTTPAPI.request` method.

        Returns:
            The ID of the event that was sent.
        """
        return self.send_text(room_id, text, html, MessageType.NOTICE, relates_to, **kwargs)

    def send_emote(self, room_id: RoomID, text: str, html: Optional[str] = None,
                   relates_to: Optional[RelatesTo] = None, **kwargs) -> Awaitable[EventID]:
        """
        Send an emote to a room. Emotes are usually displayed by prepending a star and the user's
        display name to the message, which means they're usually written in the third person.

        Args:
            room_id: The ID of the room to send the message to.
            text: The text to send. If set to None, the given HTML is used instead.
            html: The HTML to send.
            msgtype: The message type to send.
                Defaults to :attr:`MessageType.TEXT` (normal text message)
            relates_to: Message relation metadata used for things like replies.
            **kwargs: Optional parameters to pass to the :meth:`HTTPAPI.request` method.

        Returns:
            The ID of the event that was sent.
        """
        return self.send_text(room_id, text, html, MessageType.EMOTE, relates_to, **kwargs)

    def send_file(self, room_id: RoomID, url: ContentURI, info: Optional[BaseFileInfo] = None,
                  file_name: str = None, file_type: MessageType = MessageType.FILE,
                  relates_to: Optional[RelatesTo] = None, **kwargs) -> Awaitable[EventID]:
        """
        Send a file to a room.

        Args:
            room_id: The ID of the room to send the message to.
            url: The Matrix content repository URI of the file. You can upload files using
                :meth:`~MediaRepositoryMethods.upload_media`.
            info: Additional metadata about the file, e.g. mimetype, image size, video duration, etc
            file_name: The name for the file to send.
            file_type: The general file type to send. The file type can be further specified by
                setting the ``mimetype`` field of the ``info`` parameter. Defaults to
                :attr:`MessageType.FILE` (unspecified file type, e.g. document)
            relates_to: Message relation metadata used for things like replies.
            **kwargs: Optional parameters to pass to the :meth:`HTTPAPI.request` method.

        Returns:
            The ID of the event that was sent.
        """
        return self.send_message(room_id,
                                 MediaMessageEventContent(url=url, info=info, body=file_name,
                                                          relates_to=relates_to,
                                                          msgtype=file_type), **kwargs)

    def send_sticker(self, room_id: RoomID, url: ContentURI, info: Optional[ImageInfo],
                     text: Optional[str] = "", relates_to: Optional[RelatesTo] = None, **kwargs
                     ) -> Awaitable[EventID]:
        """
        Send a sticker to a room. Stickers are basically images, but they're usually rendered
        slightly differently.

        Args:
            room_id: The ID of the room to send the message to.
            url: The Matrix content repository URI of the sticker. You can upload files using
                :meth:`~MediaRepositoryMethods.upload_media`.
            info: Additional metadata about the sticker, e.g. mimetype and image size
            text: A textual description of the sticker.
            relates_to: Message relation metadata used for things like replies.
            **kwargs: Optional parameters to pass to the :meth:`HTTPAPI.request` method.

        Returns:
            The ID of the event that was sent.
        """
        return self.send_message_event(room_id, EventType.STICKER,
                                       MediaMessageEventContent(url=url, info=info, body=text,
                                                                relates_to=relates_to),
                                       **kwargs)

    def send_image(self, room_id: RoomID, url: ContentURI, info: Optional[ImageInfo] = None,
                   file_name: str = None, relates_to: Optional[RelatesTo] = None, **kwargs
                   ) -> Awaitable[EventID]:
        """
        Send an image to a room.

        Args:
            room_id: The ID of the room to send the message to.
            url: The Matrix content repository URI of the image. You can upload files using
                :meth:`~MediaRepositoryMethods.upload_media`.
            info: Additional metadata about the image, e.g. mimetype and image size
            file_name: The file name for the image to send.
            relates_to: Message relation metadata used for things like replies.
            **kwargs: Optional parameters to pass to the :meth:`HTTPAPI.request` method.

        Returns:
            The ID of the event that was sent.
        """
        return self.send_file(room_id, url, info, file_name, MessageType.IMAGE, relates_to,
                              **kwargs)

    # endregion

    # endregion
    # region 8.5 Redactions
    # API reference: https://matrix.org/docs/spec/client_server/r0.4.0.html#redactions

    async def redact(self, room_id: RoomID, event_id: EventID, reason: Optional[str] = "", **kwargs
                     ) -> EventID:
        """
        Send an event to redact a previous event.

        Redacting an event strips all information out of an event which isn't critical to the
        integrity of the server-side representation of the room.

        This cannot be undone.

        Users may redact their own events, and any user with a power level greater than or equal to
        the redact power level of the room may redact events there.

        See also: `/redact API reference`_

        Args:
            room_id: The ID of the room the event is in.
            event_id: The ID of the event to redact.
            reason: The reason for the event being redacted.
            **kwargs: Optional parameters to pass to the :meth:`HTTPAPI.request` method. Used by
                :class:`IntentAPI` to pass the timestamp massaging field to
                :meth:`AppServiceAPI.request`.

        Returns:
            The ID of the event that was sent to redact the other event.

        .. _/redact API reference:
            https://matrix.org/docs/spec/client_server/r0.5.0#put-matrix-client-r0-rooms-roomid-redact-eventid-txnid
        """
        url = Path.rooms[room_id].redact[event_id][self.api.get_txn_id()]
        resp = await self.api.request(Method.PUT, url, content={"reason": reason}, **kwargs)
        try:
            return resp["event_id"]
        except KeyError:
            raise MatrixResponseError("`event_id` not in response.")

    # endregion



# Copyright (c) 2020 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from typing import Optional

from mautrix.types import (UserID, EventType, RequestedKeyInfo, RoomKeyWithheldCode,
                           RoomKeyWithheldEventContent, RoomKeyRequestEventContent, ToDeviceEvent,
                           KeyRequestAction, ForwardedRoomKeyEventContent, EncryptionAlgorithm)
from mautrix.errors import MatrixError, MatrixConnectionError, MatrixRequestError

from .device_lists import DeviceListMachine
from .encrypt_olm import OlmEncryptionMachine
from .types import DeviceIdentity, TrustState


class RejectKeyShare(MatrixError):
    def __init__(self, log_message: str = "", code: Optional[RoomKeyWithheldCode] = None,
                 reason: Optional[str] = None) -> None:
        """
        RejectKeyShare is an error used to signal that a key share request should be rejected.

        Args:
            log_message: The message to log when rejecting the request.
            code: The m.room_key.withheld code, or ``None`` to reject silently.
            reason: The human-readable reason for the rejection.
        """
        super().__init__(log_message)
        self.code = code
        self.reason = reason


class KeySharingMachine(OlmEncryptionMachine, DeviceListMachine):
    async def default_allow_key_share(self, device: DeviceIdentity, request: RequestedKeyInfo
                                      ) -> bool:
        """
        Check whether or not the given key request should be fulfilled. You can set a custom
        function in :attr:`allow_key_share` to override this.

        Args:
            device: The identity of the device requesting keys.
            request: The requested key details.

        Returns:
            ``True`` if the key share should be accepted,
            ``False`` if it should be silently ignored.

        Raises:
            RejectKeyShare: if the key share should be rejected.
        """
        if device.user_id != self.client.mxid:
            raise RejectKeyShare(f"Rejecting key request from a different user ({device.user_id})",
                                 code=RoomKeyWithheldCode.UNAUTHORIZED,
                                 reason="This device does not share keys to other users")
        elif device.device_id == self.client.device_id:
            raise RejectKeyShare("Ignoring key request from ourselves", code=None)
        elif device.trust == TrustState.BLACKLISTED:
            raise RejectKeyShare(
                f"Rejecting key request from blacklisted device {device.device_id}",
                code=RoomKeyWithheldCode.BLACKLISTED,
                reason="You have been blacklisted by this device")
        elif device.trust == TrustState.VERIFIED:
            self.log.debug(f"Accepting key request from verified device {device.device_id}")
            return True
        elif self.share_to_unverified_devices:
            self.log.debug(f"Accepting key request from unverified device {device.device_id}, "
                           f"as share_to_unverified_devices is True")
            return True
        else:
            raise RejectKeyShare(f"Rejecting key request from unverified device {device.device_id}",
                                 code=RoomKeyWithheldCode.UNVERIFIED,
                                 reason="You have not been verified by this device")

    async def handle_room_key_request(self, evt: ToDeviceEvent, raise_exceptions: bool = False
                                      ) -> None:
        """
        Handle a ``m.room_key_request`` where the action is ``request``.

        This is automatically registered as an event handler and therefore called if the client you
        passed to the OlmMachine is syncing. You shouldn't need to call this yourself unless you
        do syncing in some manual way.

        Args:
            evt: The to-device event.
            raise_exceptions: Whether or not errors while handling should be raised.
        """
        request: RoomKeyRequestEventContent = evt.content
        if request.action != KeyRequestAction.REQUEST:
            return
        elif evt.sender == self.client.mxid and request.requesting_device_id == self.client.device_id:
            self.log.debug(f"Ignoring key request {request.request_id} from ourselves")
            return

        try:
            device = await self.get_or_fetch_device(evt.sender, request.requesting_device_id)
        except Exception:
            self.log.warning(f"Failed to get device {evt.sender}/{request.requesting_device_id} to "
                             f"handle key request {request.request_id}")
            if raise_exceptions:
                raise
            return
        if not device:
            self.log.warning(f"Couldn't find device {evt.sender}/{request.requesting_device_id} to "
                             f"handle key request {request.request_id}")
            return

        self.log.debug(f"Received key request {request.request_id} from {device.user_id}/"
                       f"{device.device_id} for session {request.body.session_id}")
        try:
            await self._handle_room_key_request(device, request.body)
        except RejectKeyShare as e:
            self.log.debug(f"Rejecting key request {request.request_id}: {e}")
            await self._reject_key_request(e, device, request.body)
        except (MatrixRequestError, MatrixConnectionError):
            self.log.exception(f"API error while handling key request {request.request_id} "
                               f"(not sending rejection)")
            if raise_exceptions:
                raise
        except Exception:
            self.log.exception(f"Error while handling key request {request.request_id}, "
                               f"sending rejection...")
            error = RejectKeyShare(code=RoomKeyWithheldCode.UNAVAILABLE,
                                   reason="An internal error occurred while trying to "
                                          "share the requested session")
            await self._reject_key_request(error, device, request.body)
            if raise_exceptions:
                raise

    async def _handle_room_key_request(self, device: DeviceIdentity, request: RequestedKeyInfo
                                       ) -> None:
        if not await self.allow_key_share(device, request):
            return

        sess = await self.crypto_store.get_group_session(request.room_id, request.sender_key,
                                                         request.session_id)
        if sess is None:
            raise RejectKeyShare(f"Didn't find group session {request.session_id} to forward to "
                                 f"{device.user_id}/{device.device_id}",
                                 code=RoomKeyWithheldCode.UNAVAILABLE,
                                 reason="Requested session ID not found on this device")

        exported_key = sess.export_session(sess.first_known_index)
        forward_content = ForwardedRoomKeyEventContent(algorithm=EncryptionAlgorithm.MEGOLM_V1,
                                                       room_id=sess.room_id, session_id=sess.id,
                                                       session_key=exported_key,
                                                       sender_key=sess.sender_key,
                                                       forwarding_key_chain=sess.forwarding_chain,
                                                       signing_key=sess.signing_key)
        await self.send_encrypted_to_device(device, EventType.FORWARDED_ROOM_KEY, forward_content)

    async def _reject_key_request(self, rejection: RejectKeyShare, device: DeviceIdentity,
                                  request: RequestedKeyInfo) -> None:
        if not rejection.code:
            # Silent rejection
            return
        content = RoomKeyWithheldEventContent(room_id=request.room_id,
                                              algorithm=request.algorithm,
                                              session_id=request.session_id,
                                              sender_key=request.sender_key,
                                              code=rejection.code,
                                              reason=rejection.reason)

        try:
            await self.client.send_to_one_device(EventType.ROOM_KEY_WITHHELD,
                                                 device.user_id, device.device_id, content)
            await self.client.send_to_one_device(EventType.ORG_MATRIX_ROOM_KEY_WITHHELD,
                                                 device.user_id, device.device_id, content)
        except MatrixError:
            self.log.warning(f"Failed to send key share rejection {rejection.code} "
                             f"to {device.user_id}/{device.device_id}", exc_info=True)



# Copyright (c) 2020 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from typing import Optional

import olm

from mautrix.types import (EncryptionAlgorithm, ToDeviceEvent, UserID, IdentityKey, OlmCiphertext,
                           OlmMsgType, EncryptedOlmEventContent)
from mautrix.errors import DecryptionError, MatchingSessionDecryptionError

from .base import BaseOlmMachine
from .types import DecryptedOlmEvent
from .sessions import Session


class OlmDecryptionMachine(BaseOlmMachine):
    async def _decrypt_olm_event(self, evt: ToDeviceEvent) -> DecryptedOlmEvent:
        if not isinstance(evt.content, EncryptedOlmEventContent):
            raise DecryptionError("unsupported event content class")
        elif evt.content.algorithm != EncryptionAlgorithm.OLM_V1:
            raise DecryptionError("unsupported event encryption algorithm")
        try:
            own_content = evt.content.ciphertext[self.account.identity_key]
        except KeyError:
            raise DecryptionError("olm event doesn't contain ciphertext for this device")

        plaintext = await self._decrypt_olm_ciphertext(evt.sender, evt.content.sender_key,
                                                       own_content)

        try:
            decrypted_evt: DecryptedOlmEvent = DecryptedOlmEvent.parse_json(plaintext)
        except Exception:
            self.log.trace("Failed to parse olm event plaintext: %s", plaintext)
            raise
        if decrypted_evt.sender != evt.sender:
            raise DecryptionError("mismatched sender in olm payload")
        elif decrypted_evt.recipient != self.client.mxid:
            raise DecryptionError("mismatched recipient in olm payload")
        elif decrypted_evt.recipient_keys.ed25519 != self.account.signing_key:
            raise DecryptionError("mismatched recipient key in olm payload")
        decrypted_evt.sender_key = evt.content.sender_key
        decrypted_evt.source = evt
        return decrypted_evt

    async def _decrypt_olm_ciphertext(self, sender: UserID, sender_key: IdentityKey,
                                      message: OlmCiphertext) -> str:
        if message.type not in (OlmMsgType.PREKEY, OlmMsgType.MESSAGE):
            raise DecryptionError("unsupported olm message type")

        try:
            plaintext = await self._try_decrypt_olm_ciphertext(sender_key, message)
        except MatchingSessionDecryptionError:
            self.log.warning(f"Found matching session yet decryption failed for sender {sender}"
                             f" with key {sender_key}")
            # TODO session unwedging
            raise

        if not plaintext:
            if message.type != OlmMsgType.PREKEY:
                # TODO session unwedging
                raise DecryptionError("Decryption failed for normal message")

            self.log.trace(f"Trying to create inbound session for {sender}/{sender_key}")
            try:
                session = await self._create_inbound_session(sender_key, message.body)
            except olm.OlmSessionError as e:
                # TODO session unwedging
                raise DecryptionError("Failed to create new session from prekey message") from e
            self.log.debug(f"Created inbound session {session.id} for {sender} "
                           f"(sender key: {sender_key})")

            try:
                plaintext = session.decrypt(message)
            except olm.OlmSessionError as e:
                raise DecryptionError("Failed to decrypt olm event with session created from "
                                      "prekey message") from e

            await self.crypto_store.update_session(sender_key, session)

        return plaintext

    async def _try_decrypt_olm_ciphertext(self, sender_key: IdentityKey, message: OlmCiphertext
                                          ) -> Optional[str]:
        sessions = await self.crypto_store.get_sessions(sender_key)
        for session in sessions:
            if message.type == OlmMsgType.PREKEY and not session.matches(message.body):
                continue

            try:
                plaintext = session.decrypt(message)
            except olm.OlmSessionError as e:
                if message.type == OlmMsgType.PREKEY:
                    raise MatchingSessionDecryptionError("decryption failed with matching session"
                                                         ) from e
            else:
                await self.crypto_store.update_session(sender_key, session)
                return plaintext
        return None

    async def _create_inbound_session(self, sender_key: IdentityKey, ciphertext: str) -> Session:
        session = self.account.new_inbound_session(sender_key, ciphertext)
        await self.crypto_store.put_account(self.account)
        await self.crypto_store.add_session(sender_key, session)
        return session



# Copyright (c) 2020 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from typing import Union, Dict, Optional, Any, List

from .abstract import StateStore
from mautrix.types import (Member, MemberStateEventContent, Membership, PowerLevelStateEventContent,
                           RoomEncryptionStateEventContent, RoomID, UserID)


class MemoryStateStore(StateStore):
    members: Dict[RoomID, Dict[UserID, Member]]
    full_member_list: Dict[RoomID, bool]
    power_levels: Dict[RoomID, PowerLevelStateEventContent]
    encryption: Dict[RoomID, Optional[RoomEncryptionStateEventContent]]

    def __init__(self) -> None:
        self.members = {}
        self.full_member_list = {}
        self.power_levels = {}
        self.encryption = {}

    def serialize(self) -> Dict[str, Any]:
        return {
            "members": {room_id: {user_id: member.serialize()
                                  for user_id, member in members.items()}
                        for room_id, members in self.members.items()},
            "full_member_list": self.full_member_list,
            "power_levels": {room_id: content.serialize()
                             for room_id, content in self.power_levels.items()},
            "encryption": {room_id: (content.serialize() if content is not None else None)
                           for room_id, content in self.encryption.items()},
        }

    def deserialize(self, data: Dict[str, Any]) -> None:
        self.members = {room_id: {user_id: Member.deserialize(member)
                                  for user_id, member in members.items()}
                        for room_id, members in data["members"].items()}
        self.full_member_list = data["full_member_list"]
        self.power_levels = {room_id: PowerLevelStateEventContent.deserialize(content)
                             for room_id, content in data["power_levels"].items()}
        self.encryption = {room_id: (RoomEncryptionStateEventContent.deserialize(content)
                                     if content is not None else None)
                           for room_id, content in data["encryption"].items()}

    async def get_member(self, room_id: RoomID, user_id: UserID) -> Optional[Member]:
        try:
            return self.members[room_id][user_id]
        except KeyError:
            return None

    async def set_member(self, room_id: RoomID, user_id: UserID,
                         member: Union[Member, MemberStateEventContent]) -> None:
        if not isinstance(member, Member):
            member = Member(membership=member.membership, avatar_url=member.avatar_url,
                            displayname=member.displayname)
        try:
            self.members[room_id][user_id] = member
        except KeyError:
            self.members[room_id] = {user_id: member}

    async def set_membership(self, room_id: RoomID, user_id: UserID,
                             membership: Membership) -> None:
        try:
            room_members = self.members[room_id]
        except KeyError:
            self.members[room_id] = {user_id: Member(membership=membership)}
            return
        try:
            room_members[user_id].membership = membership
        except (KeyError, TypeError):
            room_id[user_id] = Member(membership=membership)

    async def get_members(self, room_id: RoomID) -> Optional[List[UserID]]:
        try:
            return [user_id for user_id, member in self.members[room_id]
                    if (member.membership == Membership.JOIN
                        or member.membership == Membership.INVITE)]
        except KeyError:
            return None

    async def set_members(self, room_id: RoomID,
                          members: Dict[UserID, Union[Member, MemberStateEventContent]],
                          joined_only: bool = False) -> None:
        # TODO if joined_only is true, don't forget invited/left members
        self.members[room_id] = {user_id: (member if isinstance(member, Member)
                                           else Member(membership=member.membership,
                                                       avatar_url=member.avatar_url,
                                                       displayname=member.displayname))
                                 for user_id, member in members.items()}
        self.full_member_list[room_id] = True

    async def has_full_member_list(self, room_id: RoomID) -> bool:
        return self.full_member_list.get(room_id, False)

    async def has_power_levels_cached(self, room_id: RoomID) -> bool:
        return room_id in self.power_levels

    async def get_power_levels(self, room_id: RoomID) -> Optional[PowerLevelStateEventContent]:
        return self.power_levels.get(room_id)

    async def set_power_levels(self, room_id: RoomID, content: PowerLevelStateEventContent) -> None:
        self.power_levels[room_id] = content

    async def has_encryption_info_cached(self, room_id: RoomID) -> bool:
        return room_id in self.encryption

    async def is_encrypted(self, room_id: RoomID) -> Optional[bool]:
        try:
            return self.encryption[room_id] is not None
        except KeyError:
            return None

    async def get_encryption_info(self, room_id: RoomID) -> RoomEncryptionStateEventContent:
        return self.encryption.get(room_id)

    async def set_encryption_info(self, room_id: RoomID,
                                  content: RoomEncryptionStateEventContent) -> None:
        self.encryption[room_id] = content



# Copyright (c) 2020 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from typing import cast, Optional, Dict, Any
from datetime import datetime

import olm

from mautrix.types import (UserID, DeviceID, IdentityKey, SigningKey,
                           EncryptionKeyAlgorithm, EncryptionAlgorithm)

from .sessions import Session
from .base import canonical_json


class OlmAccount(olm.Account):
    shared: bool
    _signing_key: Optional[SigningKey]
    _identity_key: Optional[IdentityKey]

    def __init__(self) -> None:
        super().__init__()
        self.shared = False
        self._signing_key = None
        self._identity_key = None

    @property
    def signing_key(self) -> SigningKey:
        if self._signing_key is None:
            self._signing_key = self.identity_keys["ed25519"]
        return self._signing_key

    @property
    def identity_key(self) -> IdentityKey:
        if self._identity_key is None:
            self._identity_key = self.identity_keys["curve25519"]
        return self._identity_key

    @property
    def fingerprint(self) -> str:
        """
        Fingerprint is the base64-encoded signing key of this account, with spaces every 4
        characters. This is what is used for manual device verification.
        """
        key = self.signing_key
        return " ".join([key[i:i + 4] for i in range(0, len(key), 4)])

    @classmethod
    def from_pickle(cls, pickle: bytes, passphrase: str, shared: bool) -> 'OlmAccount':
        account = cast(OlmAccount, super().from_pickle(pickle, passphrase))
        account.shared = shared
        account._signing_key = None
        account._identity_key = None
        return account

    def new_inbound_session(self, sender_key: IdentityKey, ciphertext: str) -> Session:
        session = olm.InboundSession(self, olm.OlmPreKeyMessage(ciphertext), sender_key)
        self.remove_one_time_keys(session)
        return Session.from_pickle(session.pickle("roundtrip"), passphrase="roundtrip",
                                   creation_time=datetime.now())

    def new_outbound_session(self, target_key: IdentityKey, one_time_key: IdentityKey) -> Session:
        session = olm.OutboundSession(self, target_key, one_time_key)
        return Session.from_pickle(session.pickle("roundtrip"), passphrase="roundtrip",
                                   creation_time=datetime.now())

    def get_device_keys(self, user_id: UserID, device_id: DeviceID) -> Dict[str, Any]:
        device_keys = {
            "user_id": user_id,
            "device_id": device_id,
            "algorithms": [EncryptionAlgorithm.OLM_V1.value, EncryptionAlgorithm.MEGOLM_V1.value],
            "keys": {
                f"{algorithm}:{device_id}": key
                for algorithm, key in self.identity_keys.items()
            },
        }
        signature = self.sign(canonical_json(device_keys))
        device_keys["signatures"] = {
            user_id: {f"{EncryptionKeyAlgorithm.ED25519}:{device_id}": signature}
        }
        return device_keys

    def get_one_time_keys(self, user_id: UserID, device_id: DeviceID, current_otk_count: int
                          ) -> Dict[str, Any]:
        new_count = self.max_one_time_keys // 2 - current_otk_count
        if new_count > 0:
            self.generate_one_time_keys(new_count)
        keys = {}
        for key_id, key in self.one_time_keys.get("curve25519", {}).items():
            signature = self.sign(canonical_json({"key": key}))
            keys[f"{EncryptionKeyAlgorithm.SIGNED_CURVE25519}:{key_id}"] = {
                "key": key,
                "signatures": {
                    user_id: {f"{EncryptionKeyAlgorithm.ED25519}:{device_id}": signature}
                }
            }
        self.mark_keys_as_published()
        return keys



# Copyright (c) 2020 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from typing import Optional, NewType, Tuple

from mautrix.api import Path, Method
from mautrix.appservice import AppService, IntentAPI
from mautrix.types import UserID, RoomID, ContentURI
from mautrix.errors import MatrixStandardRequestError

CommunityID = NewType("CommunityID", str)


class CommunityHelper:
    az: AppService

    def __init__(self, az: AppService) -> None:
        self.az = az

    async def create(self, localpart: str) -> Tuple[CommunityID, bool]:
        try:
            resp = await self.az.intent.api.request(Method.POST, Path.create_group, {
                "localpart": localpart
            })
            return CommunityID(resp["group_id"]), True
        except MatrixStandardRequestError as e:
            if e.message == "Group already exists":
                return CommunityID(f"+{localpart}:{self.az.domain}"), False
            raise

    async def update(self, community_id: CommunityID, name: Optional[str] = None,
                     avatar_url: Optional[ContentURI] = None, short_desc: Optional[str] = None,
                     long_desc: Optional[str] = None) -> None:
        if not community_id:
            return
        await self.az.intent.api.request(Method.POST, Path.groups[community_id].profile,
                                         {k: v for k, v in {
                                             "name": name,
                                             "avatar_url": avatar_url,
                                             "short_description": short_desc,
                                             "long_description": long_desc,
                                         }.items() if v is not None})

    async def invite(self, community_id: CommunityID, user_id: UserID) -> None:
        if not community_id or not user_id:
            return
        try:
            await self.az.intent.api.request(Method.PUT,
                                             Path.groups[community_id].admin.users.invite[user_id])
        except MatrixStandardRequestError as e:
            self.az.intent.log.warning(f"Failed to invite {user_id} to {community_id}: {e}")

    async def join(self, community_id: CommunityID, intent: IntentAPI) -> bool:
        if not community_id or not intent:
            return False
        await intent.ensure_registered()
        await self.invite(community_id, intent.mxid)
        try:
            await intent.api.request(Method.PUT, Path.groups[community_id].self.accept_invite,
                                     {"m.visibility": {"type": "private"}})
        except MatrixStandardRequestError as e:
            intent.log.warning(f"Failed to join {community_id}: {e}")
        return True

    async def add_room(self, community_id: CommunityID, room_id: RoomID) -> bool:
        if not community_id or not room_id:
            return False
        try:
            await self.az.intent.api.request(Method.PUT,
                                             Path.groups[community_id].admin.rooms[room_id],
                                             {"m.visibility": {"type": "private"}})
        except MatrixStandardRequestError as e:
            self.az.intent.log.warning(f"Failed to add {room_id} to {community_id}: {e}")
        return True



# Copyright (c) 2020 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from typing import Generic, TypeVar, Any
from abc import ABC, abstractmethod
from enum import Enum
import json

from ..primitive import JSON

T = TypeVar("T")


class Serializable:
    """Serializable is the base class for types with custom JSON serializers."""

    def serialize(self) -> JSON:
        """Convert this object into objects directly serializable with `json`."""
        raise NotImplementedError()

    @classmethod
    def deserialize(cls, raw: JSON) -> Any:
        """Convert the given data parsed from JSON into an object of this type."""
        raise NotImplementedError()


class SerializerError(Exception):
    """
    SerializerErrors are raised if something goes wrong during serialization or deserialization.
    """
    pass


class GenericSerializable(ABC, Generic[T], Serializable):
    """
    An abstract Serializable that adds ``@abstractmethod`` decorators and a `Generic[T]` base class.
    """
    @classmethod
    @abstractmethod
    def deserialize(cls, raw: JSON) -> T:
        pass

    @abstractmethod
    def serialize(self) -> JSON:
        pass

    def json(self) -> str:
        """Serialize this object and dump the output as JSON."""
        return json.dumps(self.serialize())

    @classmethod
    def parse_json(cls, data: str) -> T:
        """Parse the given string as JSON and deserialize the result into this type."""
        return cls.deserialize(json.loads(data))


SerializableEnumChild = TypeVar("SerializableEnumChild", bound='SerializableEnum')


class SerializableEnum(Serializable, Enum):
    """
    A simple Serializable implementation for Enums.

    Examples:
        >>> class MyEnum(SerializableEnum):
        ...     FOO = "foo value"
        ...     BAR = "hmm"
        >>> MyEnum.FOO.serialize()
        "foo value"
        >>> MyEnum.BAR.json()
        '"hmm"'
    """

    def __init__(self, _) -> None:
        """
        A fake ``__init__`` to stop the type checker from complaining.
        Enum's ``__new__`` overrides this.
        """
        super().__init__()

    def serialize(self) -> str:
        """
        Convert this object into objects directly serializable with `json`, i.e. return the value
        set to this enum value.
        """
        return self.value

    @classmethod
    def deserialize(cls, raw: str) -> SerializableEnumChild:
        """
        Convert the given data parsed from JSON into an object of this type, i.e. find the enum
        value for the given string using ``cls(raw)``.
        """
        try:
            return cls(raw)
        except ValueError as e:
            raise SerializerError() from e

    def json(self) -> str:
        return json.dumps(self.serialize())

    @classmethod
    def parse_json(cls, data: str) -> SerializableEnumChild:
        return cls.deserialize(json.loads(data))

    def __str__(self):
        return self.value

    def __repr__(self):
        return self.value
